import xml.etree.ElementTree as Et

import pytest

from asyncua import ua
import asyncua.common.type_dictionary_builder
from asyncua.common.type_dictionary_builder import OPCTypeDictionaryBuilder, DataTypeDictionaryBuilder
from asyncua.common.type_dictionary_builder import get_ua_class, StructNode

port_num = 48540
ns_urn = 'http://test.freeopcua.github.io'


pytestmark = pytest.mark.asyncio


def to_camel_case(name):
    func = getattr(asyncua.common.type_dictionary_builder, '_to_camel_case')
    return func(name)


def reference_generator(source_id, target_id, reference_type, is_forward=True):
    func = getattr(asyncua.common.type_dictionary_builder, '_reference_generator')
    return func(source_id, target_id, reference_type, is_forward)


def set_up_test_tree():
    ext_head_attributes = {'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance', 'xmlns:tns': ns_urn,
                           'DefaultByteOrder': 'LittleEndian', 'xmlns:opc': 'http://opcfoundation.org/BinarySchema/',
                           'xmlns:ua': 'http://opcfoundation.org/UA/', 'TargetNamespace': ns_urn}

    test_etree = Et.ElementTree(Et.Element('opc:TypeDictionary', ext_head_attributes))
    name_space = Et.SubElement(test_etree.getroot(), 'opc:Import')
    name_space.attrib['Namespace'] = 'http://opcfoundation.org/UA/'
    return test_etree


@pytest.fixture(scope="function")
async def _srv(server):
    class Srv:
        pass
    srv = Srv()
    srv.srv = server
    srv.idx = await srv.srv.register_namespace(ns_urn)
    yield srv


@pytest.fixture
async def srv(_srv):
    _srv.test_etree = set_up_test_tree()
    _srv.opc_type_builder = OPCTypeDictionaryBuilder(ns_urn)
    _srv.dict_builder = DataTypeDictionaryBuilder(_srv.srv, _srv.idx, ns_urn, 'TestDict')
    await _srv.dict_builder.init()
    yield _srv


async def test_camel_case_1():
    case = 'TurtleActionlibShapeActionFeedback'
    result = to_camel_case('turtle_actionlib/ShapeActionFeedback')
    assert result == case


async def test_camel_case_2():
    case = 'HelloWorldFffD'
    result = to_camel_case('Hello#world+fff_**?&&d')
    assert result == case


async def test_opc_type_dict_process_type_opc(srv):
    case = 'opc:Boolean'
    result = getattr(srv.opc_type_builder, '_process_type')('Boolean')
    assert result == case


async def test_opc_type_dict_process_type_tns(srv):
    case = 'tns:CustomizedStruct'
    result = getattr(srv.opc_type_builder, '_process_type')('CustomizedStruct')
    assert result == case


async def test_opc_type_dict_append_struct_1(srv):
    case = {'BaseType': 'ua:ExtensionObject',
            'Name': 'CustomizedStruct'}
    result = srv.opc_type_builder.append_struct('CustomizedStruct')
    assert result.attrib == case


@pytest.mark.skip("Support for that feature currently removed")
async def test_opc_type_dict_append_struct_2(srv):
    case = {'BaseType': 'ua:ExtensionObject',
            'Name': 'CustomizedStruct'}
    result = srv.opc_type_builder.append_struct('customized_#?+`struct')
    assert result.attrib == case


async def test_opc_type_dict_add_field_1(srv):
    structure_name = 'CustomizedStruct'
    srv.opc_type_builder.append_struct(structure_name)
    srv.opc_type_builder.add_field(ua.VariantType.Boolean, 'id', structure_name)
    case = {'TypeName': 'opc:Boolean',
            'Name': 'id'}
    struct_dict = getattr(srv.opc_type_builder, '_structs_dict')
    result = list(struct_dict[structure_name])[0]
    assert result.attrib == case


async def test_opc_type_dict_add_field_2(srv):
    structure_name = 'CustomizedStruct'
    srv.opc_type_builder.append_struct(structure_name)
    srv.opc_type_builder.add_field('Boolean', 'id', structure_name)
    case = {'TypeName': 'opc:Boolean',
            'Name': 'id'}
    struct_dict = getattr(srv.opc_type_builder, '_structs_dict')
    result = list(struct_dict[structure_name])[0]
    assert result.attrib == case


async def test_opc_type_dict_add_field_3(srv):
    structure_name = 'CustomizedStruct'
    srv.opc_type_builder.append_struct(structure_name)
    srv.opc_type_builder.add_field(ua.VariantType.Boolean, 'id', structure_name, is_array=True)
    case = [{'TypeName': 'opc:Int32',
            'Name': 'NoOfid'},
            {'TypeName': 'opc:Boolean',
             'LengthField': 'NoOfid',
             'Name': 'id'}]
    struct_dict = getattr(srv.opc_type_builder, '_structs_dict')
    result = [item.attrib for item in list(struct_dict[structure_name])]
    assert result == case


async def test_opc_type_dict_get_dict_value(srv):
    structure_name = 'CustomizedStruct'
    srv.opc_type_builder.append_struct(structure_name)
    # external tree operation
    appended_struct = Et.SubElement(srv.test_etree.getroot(), 'opc:StructuredType')
    appended_struct.attrib['BaseType'] = 'ua:ExtensionObject'
    appended_struct.attrib['Name'] = to_camel_case(structure_name)

    srv.opc_type_builder.add_field(ua.VariantType.Boolean, 'id', structure_name)
    # external tree operation
    field = Et.SubElement(appended_struct, 'opc:Field')
    field.attrib['Name'] = 'id'
    field.attrib['TypeName'] = 'opc:Boolean'
    case = Et.tostring(srv.test_etree.getroot(), encoding='utf-8').decode("utf-8").replace(' ', '')
    result = srv.opc_type_builder.get_dict_value().decode("utf-8").replace(' ', '').replace('\n', '')
    assert result == case


async def test_reference_generator_1(srv):
    id1 = ua.NodeId(1, NamespaceIndex=2, NodeIdType=ua.NodeIdType.Numeric)
    id2 = ua.NodeId(2, NamespaceIndex=2, NodeIdType=ua.NodeIdType.Numeric)
    ref = ua.NodeId(ua.ObjectIds.HasEncoding, 0)
    result = reference_generator(id1, id2, ref)
    assert result.IsForward
    assert result.ReferenceTypeId == ref
    assert result.SourceNodeId == id1
    assert result.TargetNodeClass == ua.NodeClass.DataType
    assert result.TargetNodeId == id2


async def test_reference_generator_2(srv):
    id1 = ua.NodeId(1, NamespaceIndex=2, NodeIdType=ua.NodeIdType.Numeric)
    id2 = ua.NodeId(2, NamespaceIndex=2, NodeIdType=ua.NodeIdType.Numeric)
    ref = ua.NodeId(ua.ObjectIds.HasEncoding, 0)
    result = reference_generator(id1, id2, ref, False)
    assert not result.IsForward
    assert result.ReferenceTypeId == ref
    assert result.SourceNodeId == id1
    assert result.TargetNodeClass == ua.NodeClass.DataType
    assert result.TargetNodeId == id2


async def test_data_type_dict_general(srv):
    assert srv.dict_builder.dict_id is not None
    assert getattr(srv.dict_builder, '_type_dictionary') is not None


async def test_data_type_dict_add_dictionary(srv):
    add_dictionary = getattr(srv.dict_builder, '_add_dictionary')
    dict_name = 'TestDict'
    dict_node = srv.srv.get_node(await add_dictionary(dict_name))
    assert await dict_node.read_browse_name() == ua.QualifiedName(dict_name, srv.idx)
    assert await dict_node.read_node_class() == ua.NodeClass.Variable
    assert (await dict_node.get_parent()).nodeid == ua.NodeId(ua.ObjectIds.OPCBinarySchema_TypeSystem, 0)
    assert ua.NodeId(ua.ObjectIds.HasComponent, 0) == (await dict_node.get_references(refs=ua.ObjectIds.HasComponent))[0].ReferenceTypeId
    assert await dict_node.read_type_definition() == ua.NodeId(ua.ObjectIds.DataTypeDictionaryType, 0)
    assert await dict_node.read_display_name() == ua.LocalizedText(dict_name)
    assert await dict_node.read_data_type() == ua.NodeId(ua.ObjectIds.ByteString)
    assert await dict_node.read_value_rank() == -1


async def test_data_type_dict_create_data_type(srv):
    type_name = 'CustomizedStruct2'
    created_type = await srv.dict_builder.create_data_type(type_name)
    assert isinstance(created_type, StructNode)
    # Test data type node
    type_node = srv.srv.get_node(created_type.data_type)
    assert await type_node.read_browse_name() == ua.QualifiedName(type_name, srv.idx)
    assert await type_node.read_node_class() == ua.NodeClass.DataType
    assert (await type_node.get_parent()).nodeid == ua.NodeId(ua.ObjectIds.Structure, 0)
    assert ua.NodeId(ua.ObjectIds.HasSubtype, 0) == (await type_node.get_references(refs=ua.ObjectIds.HasSubtype))[0].ReferenceTypeId
    assert await type_node.read_display_name() == ua.LocalizedText(type_name)

    # Test description node
    n = srv.srv.get_node(srv.dict_builder.dict_id)
    desc_node = await n.get_child(f"{srv.dict_builder._idx}:{type_name}")
    assert await desc_node.read_browse_name() == ua.QualifiedName(type_name, srv.idx)
    assert await desc_node.read_node_class() == ua.NodeClass.Variable
    assert (await desc_node.get_parent()).nodeid == srv.dict_builder.dict_id
    assert ua.NodeId(ua.ObjectIds.HasComponent, 0) == (await desc_node.get_references(refs=ua.ObjectIds.HasComponent))[0].ReferenceTypeId
    assert await desc_node.read_type_definition() == ua.NodeId(ua.ObjectIds.DataTypeDescriptionType, 0)

    assert await desc_node.read_display_name() == ua.LocalizedText(type_name)
    assert await desc_node.read_data_type() == ua.NodeId(ua.ObjectIds.String)
    assert await desc_node.read_value() == type_name
    assert await desc_node.read_value_rank() == -1

    # Test object node
    obj_node = (await type_node.get_children(refs=ua.ObjectIds.HasEncoding))[0]
    assert await obj_node.read_browse_name() == ua.QualifiedName('Default Binary', 0)
    assert await obj_node.read_node_class() == ua.NodeClass.Object
    assert (await obj_node.get_references(refs=ua.ObjectIds.HasEncoding))[0].NodeId == type_node.nodeid
    assert ua.NodeId(ua.ObjectIds.HasEncoding, 0) == (await obj_node.get_references(refs=ua.ObjectIds.HasEncoding))[0].ReferenceTypeId
    assert await obj_node.read_type_definition() == ua.NodeId(ua.ObjectIds.DataTypeEncodingType, 0)
    assert await obj_node.read_display_name() == ua.LocalizedText('Default Binary')
    assert len(await obj_node.read_event_notifier()) == 0

    # Test links, three were tested above
    struct_node = srv.srv.get_node(ua.NodeId(ua.ObjectIds.Structure, 0))
    struct_children = await struct_node.get_children(refs=ua.ObjectIds.HasSubtype)
    assert type_node in struct_children
    dict_node = srv.srv.get_node(srv.dict_builder.dict_id)
    dict_children = await dict_node.get_children(refs=ua.ObjectIds.HasComponent)
    assert desc_node in dict_children
    assert obj_node in await type_node.get_children(ua.ObjectIds.HasEncoding)
    assert desc_node in await obj_node.get_children(refs=ua.ObjectIds.HasDescription)
    assert obj_node.nodeid == (await desc_node.get_references(refs=ua.ObjectIds.HasDescription, direction=ua.BrowseDirection.Inverse))[0].NodeId


async def test_data_type_dict_set_dict_byte_string(srv):
    structure_name = 'CustomizedStruct'
    await srv.dict_builder.create_data_type(structure_name)
    srv.dict_builder.add_field(ua.VariantType.Int32, 'id', structure_name)
    await srv.dict_builder.set_dict_byte_string()
    # external tree operation
    appended_struct = Et.SubElement(srv.test_etree.getroot(), 'opc:StructuredType')
    appended_struct.attrib['BaseType'] = 'ua:ExtensionObject'
    appended_struct.attrib['Name'] = to_camel_case(structure_name)

    # external tree operation
    field = Et.SubElement(appended_struct, 'opc:Field')
    field.attrib['Name'] = 'id'
    field.attrib['TypeName'] = 'opc:Int32'
    case = Et.tostring(srv.test_etree.getroot(), encoding='utf-8').decode("utf-8").replace(' ', '')
    result = (await srv.srv.get_node(srv.dict_builder.dict_id).read_value()).decode("utf-8").replace(' ', '').replace('\n', '')
    assert result == case


async def test_data_type_dict_add_field_1(srv):
    struct_name = 'CustomizedStruct'
    await srv.dict_builder.create_data_type(struct_name)
    srv.dict_builder.add_field(ua.VariantType.Int32, 'id', struct_name)
    await srv.dict_builder.set_dict_byte_string()
    await srv.srv.load_type_definitions()
    struct = get_ua_class(struct_name)
    struct_instance = struct()
    assert struct_instance.id == 0


async def test_data_type_dict_add_field_2(srv):
    struct_name = 'AnotherCustomizedStruct'
    await srv.dict_builder.create_data_type(struct_name)
    srv.dict_builder.add_field(ua.VariantType.Int32, 'id', struct_name, is_array=True)
    await srv.dict_builder.set_dict_byte_string()
    await srv.srv.load_type_definitions()
    struct = get_ua_class(struct_name)
    struct_instance = struct()
    assert isinstance(struct_instance.id, list)


async def test_struct_node_general(srv):
    struct_name = 'CustomizedStruct'
    struct_node = await srv.dict_builder.create_data_type(struct_name)
    assert getattr(struct_node, '_type_dict'), srv.dict_builder
    assert isinstance(struct_node.data_type, ua.NodeId)
    assert struct_node.name == struct_name


async def test_struct_node_add_field(srv):
    struct_name = 'CustomizedStruct'
    struct_node = await srv.dict_builder.create_data_type(struct_name)
    struct_node.add_field('id', ua.VariantType.Int32)
    await srv.dict_builder.set_dict_byte_string()
    await srv.srv.load_type_definitions()
    struct = get_ua_class(struct_name)
    struct_instance = struct()
    assert struct_instance.id == 0


async def test_get_ua_class_1(srv):
    struct_name = 'CustomizedStruct'
    struct_node = await srv.dict_builder.create_data_type(struct_name)
    struct_node.add_field('id', ua.VariantType.Int32)
    await srv.dict_builder.set_dict_byte_string()
    await srv.srv.load_type_definitions()
    try:
        assert get_ua_class(struct_name) is not None
    except AttributeError:
        pass


@pytest.mark.skip("Support for that feature currently removed")
async def test_get_ua_class_2(srv):
    struct_name = '*c*u_stom-ized&Stru#ct'
    struct_node = await srv.dict_builder.create_data_type(struct_name)
    struct_node.add_field('id', ua.VariantType.Int32)
    await srv.dict_builder.set_dict_byte_string()
    await srv.srv.load_type_definitions()
    try:
        assert get_ua_class(struct_name) is not None
    except AttributeError:
        pass


async def test_functional_basic(srv):
    basic_struct_name = 'basic_structure'
    basic_struct = await srv.dict_builder.create_data_type(basic_struct_name)
    basic_struct.add_field('ID', ua.VariantType.Int32)
    basic_struct.add_field('Gender', ua.VariantType.Boolean)
    basic_struct.add_field('Comments', ua.VariantType.String)

    await srv.dict_builder.set_dict_byte_string()
    await srv.srv.load_type_definitions()

    basic_var = await srv.srv.nodes.objects.add_variable(ua.NodeId(NamespaceIndex=srv.idx), 'BasicStruct',
                                                    ua.Variant(None, ua.VariantType.Null),
                                                    datatype=basic_struct.data_type)

    basic_msg = get_ua_class(basic_struct_name)()
    basic_msg.ID = 3
    basic_msg.Gender = True
    basic_msg.Comments = 'Test string'
    await basic_var.write_value(basic_msg)

    basic_result = await basic_var.read_value()
    assert basic_result == basic_msg
    await srv.srv.delete_nodes([basic_var])


async def test_functional_advance(srv):
    basic_struct_name = 'basic_structure'
    basic_struct = await srv.dict_builder.create_data_type(basic_struct_name)
    basic_struct.add_field('ID', ua.VariantType.Int32)
    basic_struct.add_field('Gender', ua.VariantType.Boolean)
    basic_struct.add_field('Comments', ua.VariantType.String)

    nested_struct_name = 'nested_structure'
    nested_struct = await srv.dict_builder.create_data_type(nested_struct_name)
    nested_struct.add_field('Name', ua.VariantType.String)
    nested_struct.add_field('Surname', ua.VariantType.String)
    nested_struct.add_field('Stuff', basic_struct)

    await srv.dict_builder.set_dict_byte_string()
    await srv.srv.load_type_definitions()

    basic_var = await srv.srv.nodes.objects.add_variable(ua.NodeId(NamespaceIndex=srv.idx), 'BasicStruct',
                                                    ua.Variant(None, ua.VariantType.ExtensionObject),
                                                    datatype=basic_struct.data_type)

    basic_msg = get_ua_class(basic_struct_name)()
    basic_msg.ID = 3
    basic_msg.Gender = True
    basic_msg.Comments = 'Test string'
    await basic_var.write_value(basic_msg)

    nested_var = await srv.srv.nodes.objects.add_variable(ua.NodeId(NamespaceIndex=srv.idx), 'NestedStruct',
                                                     ua.Variant(None, ua.VariantType.ExtensionObject),
                                                     datatype=nested_struct.data_type)

    nested_msg = get_ua_class(nested_struct_name)()
    nested_msg.Stuff = basic_msg
    nested_msg.Name = 'Max'
    nested_msg.Surname = 'Karl'
    await nested_var.write_value(nested_msg)

    basic_result = await basic_var.read_value()
    assert basic_result == basic_msg
    nested_result = await nested_var.read_value()
    assert nested_result == nested_msg
    await srv.srv.delete_nodes([basic_var, nested_var])



"""
Binary protocol specific functions and constants
"""

import functools
import struct
import logging
from typing import Any, Callable
import uuid
from enum import IntEnum, Enum, IntFlag
from dataclasses import is_dataclass, fields, Field
from asyncua import ua
from .uaerrors import UaError
from ..common.utils import Buffer
from .uatypes import type_is_list, type_is_union, type_from_list, type_from_union

logger = logging.getLogger('__name__')


def test_bit(data, offset):
    mask = 1 << offset
    return data & mask


def set_bit(data, offset):
    mask = 1 << offset
    return data | mask


def unset_bit(data, offset):
    mask = 1 << offset
    return data & ~mask


class _DateTime:
    @staticmethod
    def pack(dt):
        epch = ua.datetime_to_win_epoch(dt)
        return Primitives.Int64.pack(epch)

    @staticmethod
    def unpack(data):
        epch = Primitives.Int64.unpack(data)
        return ua.win_epoch_to_datetime(epch)


class _Bytes:
    @staticmethod
    def pack(data):
        if data is None:
            return Primitives.Int32.pack(-1)
        length = len(data)
        return Primitives.Int32.pack(length) + data

    @staticmethod
    def unpack(data):
        length = Primitives.Int32.unpack(data)
        if length == -1:
            return None
        return data.read(length)


class _String:
    @staticmethod
    def pack(string):
        if string is not None:
            string = string.encode('utf-8')
        return _Bytes.pack(string)

    @staticmethod
    def unpack(data):
        b = _Bytes.unpack(data)
        if b is None:
            return b
        return b.decode('utf-8', errors="replace")  # not need to be strict here, this is user data


class _Null:
    @staticmethod
    def pack(data):
        return b''

    @staticmethod
    def unpack(data):
        return None


class _Guid:
    @staticmethod
    def pack(guid):
        # convert python UUID 6 field format to OPC UA 4 field format
        f1 = Primitives.UInt32.pack(guid.time_low)
        f2 = Primitives.UInt16.pack(guid.time_mid)
        f3 = Primitives.UInt16.pack(guid.time_hi_version)
        f4a = Primitives.Byte.pack(guid.clock_seq_hi_variant)
        f4b = Primitives.Byte.pack(guid.clock_seq_low)
        f4c = struct.pack('>Q', guid.node)[2:8]  # no primitive .pack available for 6 byte int
        f4 = f4a + f4b + f4c
        # concat byte fields
        b = f1 + f2 + f3 + f4

        return b

    @staticmethod
    def unpack(data):
        # convert OPC UA 4 field format to python UUID bytes
        f1 = struct.pack('>I', Primitives.UInt32.unpack(data))
        f2 = struct.pack('>H', Primitives.UInt16.unpack(data))
        f3 = struct.pack('>H', Primitives.UInt16.unpack(data))
        f4 = data.read(8)
        # concat byte fields
        b = f1 + f2 + f3 + f4

        return uuid.UUID(bytes=b)


class _Primitive1:
    def __init__(self, fmt):
        self._fmt = fmt
        st = struct.Struct(fmt.format(1))
        self.size = st.size
        self.format = st.format

    def pack(self, data):
        return struct.pack(self.format, data)

    def unpack(self, data):
        return struct.unpack(self.format, data.read(self.size))[0]

    def pack_array(self, data):
        if data is None:
            return Primitives.Int32.pack(-1)
        if not isinstance(data, list):
            logger.warning('ua_binary.py > _Primitive1 > pack_array > data: %s is not a instance of "list"!', data)
            return Primitives.Int32.pack(-1)  # to prevent crashing while runtime
        size_data = Primitives.Int32.pack(len(data))
        return size_data + struct.pack(self._fmt.format(len(data)), *data)

    def unpack_array(self, data, length):
        if length == -1:
            return None
        if length == 0:
            return ()
        return struct.unpack(self._fmt.format(length), data.read(self.size * length))


class Primitives1:
    SByte = _Primitive1('<{:d}b')
    Int16 = _Primitive1('<{:d}h')
    Int32 = _Primitive1('<{:d}i')
    Int64 = _Primitive1('<{:d}q')
    Byte = _Primitive1('<{:d}B')
    Char = Byte
    UInt16 = _Primitive1('<{:d}H')
    UInt32 = _Primitive1('<{:d}I')
    UInt64 = _Primitive1('<{:d}Q')
    Boolean = _Primitive1('<{:d}?')
    Double = _Primitive1('<{:d}d')
    Float = _Primitive1('<{:d}f')


class Primitives(Primitives1):
    Null = _Null()
    String = _String()
    Bytes = _Bytes()
    ByteString = _Bytes()
    CharArray = _String()
    DateTime = _DateTime()
    Guid = _Guid()


@functools.lru_cache(maxsize=None)
def create_uatype_serializer(vtype):
    if hasattr(Primitives, vtype.name):
        return getattr(Primitives, vtype.name).pack
    if vtype.value > 25:
        return Primitives.Bytes.pack
    if vtype == ua.VariantType.ExtensionObject:
        return extensionobject_to_binary
    if vtype in (ua.VariantType.NodeId, ua.VariantType.ExpandedNodeId):
        return nodeid_to_binary
    if vtype == ua.VariantType.Variant:
        return variant_to_binary
    return struct_to_binary


def pack_uatype(vtype, value):
    return create_uatype_serializer(vtype)(value)


def unpack_uatype(vtype, data):
    if hasattr(Primitives, vtype.name):
        st = getattr(Primitives, vtype.name)
        return st.unpack(data)
    if vtype.value > 25:
        return Primitives.Bytes.unpack(data)
    if vtype == ua.VariantType.ExtensionObject:
        return extensionobject_from_binary(data)
    if vtype in (ua.VariantType.NodeId, ua.VariantType.ExpandedNodeId):
        return nodeid_from_binary(data)
    if vtype == ua.VariantType.Variant:
        return variant_from_binary(data)
    if hasattr(ua, vtype.name):
        cls = getattr(ua, vtype.name)
        return struct_from_binary(cls, data)
    raise UaError(f'Cannot unpack unknown variant type {vtype}')


@functools.lru_cache(maxsize=None)
def create_uatype_array_serializer(vtype):
    if hasattr(Primitives1, vtype.name):
        data_type = getattr(Primitives1, vtype.name)
        return data_type.pack_array
    serializer = create_uatype_serializer(vtype)

    def serialize(array):
        if array is None:
            return b'\xff\xff\xff\xff'
        length = Primitives.Int32.pack(len(array))
        return length + b"".join(serializer(val) for val in array)
    return serialize


def pack_uatype_array(vtype, array):
    return create_uatype_array_serializer(vtype)(array)


def unpack_uatype_array(vtype, data):
    length = Primitives.Int32.unpack(data)
    if length == -1:
        return None
    if hasattr(Primitives1, vtype.name):
        data_type = getattr(Primitives1, vtype.name)
        # Remark: works without tuple conversion to list.
        return list(data_type.unpack_array(data, length))
    # Revert to slow serial unpacking.
    return [unpack_uatype(vtype, data) for _ in range(length)]


def field_serializer(field: Field) -> Callable[[Any], bytes]:
    is_optional = type_is_union(field.type)
    uatype = field.type
    if is_optional:
        uatype = type_from_union(uatype)
    if type_is_list(uatype):
        return create_list_serializer(type_from_list(uatype))
    else:
        serializer = create_type_serializer(uatype)
        if is_optional:
            return lambda val: b'' if val is None else serializer(val)
        else:
            return serializer


@functools.lru_cache(maxsize=None)
def create_dataclass_serializer(dataclazz):
    """Given a dataclass, return a function that serializes instances of this dataclass"""
    data_fields = fields(dataclazz)
    union_fields_encodings = [  # Name and binary encoding of optional fields
        (field.name, 1 << enc_count)
        for enc_count, field
        in enumerate(filter(lambda f: type_is_union(f.type), data_fields))
    ]

    def serialize_encoding(obj):
        enc = 0
        for name, enc_val in union_fields_encodings:
            if obj.__dict__[name] is not None:
                enc |= enc_val
        return Primitives.Byte.pack(enc)

    encoding_functions = [(f.name, field_serializer(f)) for f in data_fields]

    def serialize(obj):
        return b''.join(
            serialize_encoding(obj) if name == 'Encoding'
            else serializer(obj.__dict__[name])
            for name, serializer in encoding_functions
        )

    return serialize


def struct_to_binary(obj):
    serializer = create_dataclass_serializer(obj.__class__)
    return serializer(obj)


@functools.lru_cache(maxsize=None)
def create_type_serializer(uatype):
    """Create a binary serialization function for the given UA type"""
    if type_is_list(uatype):
        return create_list_serializer(type_from_list(uatype))
    if hasattr(Primitives, uatype.__name__):
        return getattr(Primitives, uatype.__name__).pack
    if issubclass(uatype, Enum):
        return lambda val: \
            Primitives.Int32.pack(val.value) if isinstance(val, (IntEnum, Enum, IntFlag)) \
            else Primitives.Int32.pack(val)
    if hasattr(ua.VariantType, uatype.__name__):
        vtype = getattr(ua.VariantType, uatype.__name__)
        return create_uatype_serializer(vtype)
    if issubclass(uatype, ua.NodeId):
        return nodeid_to_binary
    if issubclass(uatype, ua.Variant):
        return variant_to_binary
    if is_dataclass(uatype):
        return create_dataclass_serializer(uatype)
    raise UaError(f'No known way to pack value of type {uatype} to ua binary')


def to_binary(uatype, val):
    return create_type_serializer(uatype)(val)


@functools.lru_cache(maxsize=None)
def create_list_serializer(uatype) -> Callable[[Any], bytes]:
    """
    Given a type, return a function that takes a list of instances
    of that type and serializes it.
    """
    if hasattr(Primitives1, uatype.__name__):
        data_type = getattr(Primitives1, uatype.__name__)
        return data_type.pack_array
    type_serializer = create_type_serializer(uatype)
    none_val = Primitives.Int32.pack(-1)

    def serialize(val):
        if val is None:
            return none_val
        data_size = Primitives.Int32.pack(len(val))
        return data_size + b''.join(type_serializer(el) for el in val)
    return serialize


def list_to_binary(uatype, val):
    return create_list_serializer(uatype)(val)


def nodeid_to_binary(nodeid):
    if nodeid.NodeIdType == ua.NodeIdType.TwoByte:
        data = struct.pack('<BB', nodeid.NodeIdType.value, nodeid.Identifier)
    elif nodeid.NodeIdType == ua.NodeIdType.FourByte:
        data = struct.pack('<BBH', nodeid.NodeIdType.value, nodeid.NamespaceIndex, nodeid.Identifier)
    elif nodeid.NodeIdType == ua.NodeIdType.Numeric:
        data = struct.pack('<BHI', nodeid.NodeIdType.value, nodeid.NamespaceIndex, nodeid.Identifier)
    elif nodeid.NodeIdType == ua.NodeIdType.String:
        data = struct.pack('<BH', nodeid.NodeIdType.value, nodeid.NamespaceIndex) + \
            Primitives.String.pack(nodeid.Identifier)
    elif nodeid.NodeIdType == ua.NodeIdType.ByteString:
        data = struct.pack('<BH', nodeid.NodeIdType.value, nodeid.NamespaceIndex) + \
            Primitives.Bytes.pack(nodeid.Identifier)
    elif nodeid.NodeIdType == ua.NodeIdType.Guid:
        data = struct.pack('<BH', nodeid.NodeIdType.value, nodeid.NamespaceIndex) + \
            Primitives.Guid.pack(nodeid.Identifier)
    else:
        raise UaError(f'Unknown NodeIdType: {nodeid.NodeIdType} for NodeId: {nodeid}')
    # Add NamespaceUri and ServerIndex in case we have an ExpandedNodeId
    if hasattr(nodeid, "NamespaceUri") and nodeid.NamespaceUri:
        data = bytearray(data)
        data[0] = set_bit(data[0], 7)
        data.extend(Primitives.String.pack(nodeid.NamespaceUri))
    if hasattr(nodeid, "ServerIndex") and nodeid.ServerIndex:
        if not isinstance(data, bytearray):
            data = bytearray(data)
        data[0] = set_bit(data[0], 6)
        data.extend(Primitives.UInt32.pack(nodeid.ServerIndex))
    return data


def nodeid_from_binary(data):
    encoding = ord(data.read(1))
    nidtype = ua.NodeIdType(encoding & 0b00111111)
    uri = None
    server_idx = None

    if nidtype == ua.NodeIdType.TwoByte:
        identifier = ord(data.read(1))
        nidx = 0
    elif nidtype == ua.NodeIdType.FourByte:
        nidx, identifier = struct.unpack("<BH", data.read(3))
    elif nidtype == ua.NodeIdType.Numeric:
        nidx, identifier = struct.unpack("<HI", data.read(6))
    elif nidtype == ua.NodeIdType.String:
        nidx = Primitives.UInt16.unpack(data)
        identifier = Primitives.String.unpack(data)
    elif nidtype == ua.NodeIdType.ByteString:
        nidx = Primitives.UInt16.unpack(data)
        identifier = Primitives.Bytes.unpack(data)
    elif nidtype == ua.NodeIdType.Guid:
        nidx = Primitives.UInt16.unpack(data)
        identifier = Primitives.Guid.unpack(data)
    else:
        raise UaError(f'Unknown NodeId encoding: {nidtype}')

    if test_bit(encoding, 7):
        uri = Primitives.String.unpack(data)
    if test_bit(encoding, 6):
        server_idx = Primitives.UInt32.unpack(data)

    if uri is not None or server_idx is not None:
        return ua.ExpandedNodeId(identifier, nidx, nidtype, uri, server_idx)
    return ua.NodeId(identifier, nidx, nidtype)


def variant_to_binary(var):
    b = []
    encoding = var.VariantType.value & 0b111111
    if var.is_array or isinstance(var.Value, (list, tuple)):
        encoding = set_bit(encoding, 7)
        if var.Dimensions is not None:
            encoding = set_bit(encoding, 6)
        b.append(Primitives.Byte.pack(encoding))
        b.append(pack_uatype_array(var.VariantType, ua.flatten(var.Value)))
        if var.Dimensions is not None:
            b.append(pack_uatype_array(ua.VariantType.Int32, var.Dimensions))
    else:
        b.append(Primitives.Byte.pack(encoding))
        b.append(pack_uatype(var.VariantType, var.Value))

    return b"".join(b)


def variant_from_binary(data):
    dimensions = None
    encoding = ord(data.read(1))
    int_type = encoding & 0b00111111
    vtype = ua.datatype_to_varianttype(int_type)
    if test_bit(encoding, 7):
        value = unpack_uatype_array(vtype, data)
        dimensions = [0]
    else:
        value = unpack_uatype(vtype, data)
    if test_bit(encoding, 6):
        dimensions = unpack_uatype_array(ua.VariantType.Int32, data)
        if value is not None:
            value = _reshape(value, dimensions)
    return ua.Variant(value, vtype, dimensions)


def _reshape(flat, dims):
    subdims = dims[1:]
    subsize = 1
    for i in subdims:
        if i == 0:
            i = 1
        subsize *= i
    while dims[0] * subsize > len(flat):
        flat.append([])
    if not subdims or subdims == [0]:
        return flat
    return [_reshape(flat[i:i + subsize], subdims) for i in range(0, len(flat), subsize)]


def extensionobject_from_binary(data):
    """
    Convert binary-coded ExtensionObject to a Python object.
    Returns an object, or None if TypeId is zero
    """
    typeid = nodeid_from_binary(data)
    encoding = ord(data.read(1))
    body = None
    if encoding & (1 << 0):
        length = Primitives.Int32.unpack(data)
        if length < 1:
            body = Buffer(b"")
        else:
            body = data.copy(length)
            data.skip(length)
    if typeid.Identifier == 0:
        return ua.ExtensionObject()
    if typeid in ua.extension_objects_by_typeid:
        cls = ua.extension_objects_by_typeid[typeid]
        if body is None:
            raise UaError(f'parsing ExtensionObject {cls.__name__} without data')
        return from_binary(cls, body)
    if body is not None:
        body_data = body.read(len(body))
    else:
        body_data = None
    e = ua.ExtensionObject(
        TypeId=typeid,
        Body=body_data,
    )
    return e


def extensionobject_to_binary(obj):
    """
    Convert Python object to binary-coded ExtensionObject.
    If obj is None, convert to empty ExtensionObject (TypeId=0, no Body).
    Returns a binary string
    """
    if isinstance(obj, ua.ExtensionObject):
        return struct_to_binary(obj)
    if obj is None:
        type_id = ua.NodeId()
        encoding = 0
        body = None
    else:
        type_id = ua.extension_object_typeids[obj.__class__.__name__]
        encoding = 0x01
        body = struct_to_binary(obj)
    packet = [
        nodeid_to_binary(type_id),
        Primitives.Byte.pack(encoding),
    ]
    if body:
        packet.append(Primitives.Bytes.pack(body))
    return b''.join(packet)


def from_binary(uatype, data):
    """
    unpack data given an uatype as a string or a python dataclass using ua types
    """
    if type_is_union(uatype):
        uatype = type_from_union(uatype)
    if type_is_list(uatype):
        utype = type_from_list(uatype)
        if hasattr(ua.VariantType, utype.__name__):
            vtype = getattr(ua.VariantType, utype.__name__)
            return unpack_uatype_array(vtype, data)
        size = Primitives.Int32.unpack(data)
        return [from_binary(utype, data) for _ in range(size)]
    if hasattr(ua.VariantType, uatype.__name__):
        vtype = getattr(ua.VariantType, uatype.__name__)
        return unpack_uatype(vtype, data)
    if hasattr(Primitives, uatype.__name__):
        return getattr(Primitives, uatype.__name__).unpack(data)
    return struct_from_binary(uatype, data)


def struct_from_binary(objtype, data):
    """
    unpack an ua struct. Arguments are an objtype as Python class or string
    """
    if isinstance(objtype, str):
        objtype = getattr(ua, objtype)
    if issubclass(objtype, Enum):
        return objtype(Primitives.Int32.unpack(data))
    enc_count = -1
    kwargs = {}
    enc = 0
    for field in fields(objtype):
        # if our member has a switch and it is not set we skip it
        if type_is_union(field.type):
            enc_count += 1
            if not test_bit(enc, enc_count):
                continue
        val = from_binary(field.type, data)
        if field.name == "Encoding":  # Rmq: all code written assuming encoding is called Encoding
            enc = val
        else:
            kwargs[field.name] = val
    return objtype(**kwargs)


def header_to_binary(hdr):
    b = [struct.pack("<3ss", hdr.MessageType, hdr.ChunkType)]
    size = hdr.body_size + 8
    if hdr.MessageType in (ua.MessageType.SecureOpen, ua.MessageType.SecureClose, ua.MessageType.SecureMessage):
        size += 4
    b.append(Primitives.UInt32.pack(size))
    if hdr.MessageType in (ua.MessageType.SecureOpen, ua.MessageType.SecureClose, ua.MessageType.SecureMessage):
        b.append(Primitives.UInt32.pack(hdr.ChannelId))
    return b"".join(b)


def header_from_binary(data):
    hdr = ua.Header()
    hdr.MessageType, hdr.ChunkType, hdr.packet_size = struct.unpack("<3scI", data.read(8))
    hdr.body_size = hdr.packet_size - 8
    if hdr.MessageType in (ua.MessageType.SecureOpen, ua.MessageType.SecureClose, ua.MessageType.SecureMessage):
        hdr.body_size -= 4
        hdr.ChannelId = Primitives.UInt32.unpack(data)
        hdr.header_size = 12
    return hdr


def uatcp_to_binary(message_type, message):
    """
    Convert OPC UA TCP message (see OPC UA specs Part 6, 7.1) to binary.
    The only supported types are Hello, Acknowledge and ErrorMessage
    """
    header = ua.Header(message_type, ua.ChunkType.Single)
    binmsg = struct_to_binary(message)
    header.body_size = len(binmsg)
    return header_to_binary(header) + binmsg



"""
Helper function and classes that do not rely on asyncua library.
Helper function and classes depending on ua object are in ua_utils.py
"""

import os
import logging
from ..ua.uaerrors import UaError

_logger = logging.getLogger(__name__)


class ServiceError(UaError):
    def __init__(self, code):
        super().__init__('UA Service Error')
        self.code = code


class NotEnoughData(UaError):
    pass


class SocketClosedException(UaError):
    pass


class Buffer:
    """
    Alternative to io.BytesIO making debug easier
    and added a few convenience methods.
    """

    def __init__(self, data, start_pos=0, size=-1):
        self._data = data
        self._cur_pos = start_pos
        if size == -1:
            size = len(data) - start_pos
        self._size = size

    def __str__(self):
        return f"Buffer(size:{self._size}, data:{self._data[self._cur_pos:self._cur_pos + self._size]})"
    __repr__ = __str__

    def __len__(self):
        return self._size

    def __bool__(self):
        return self._size > 0

    def __bytes__(self):
        """Return remains of buffer as bytes."""
        return self._data[self._cur_pos:]

    def read(self, size):
        """
        read and pop number of bytes for buffer
        """
        if size > self._size:
            raise NotEnoughData(f"Not enough data left in buffer, request for {size}, we have {self._size}")
        self._size -= size
        pos = self._cur_pos
        self._cur_pos += size
        return self._data[pos:self._cur_pos]

    def copy(self, size=-1):
        """
        return a shadow copy, optionally only copy 'size' bytes
        """
        if size == -1 or size > self._size:
            size = self._size
        return Buffer(self._data, self._cur_pos, size)

    def skip(self, size):
        """
        skip size bytes in buffer
        """
        if size > self._size:
            raise NotEnoughData(f"Not enough data left in buffer, request for {size}, we have {self._size}")
        self._size -= size
        self._cur_pos += size

    @property
    def cur_pos(self):
        return self._cur_pos

    def rewind(self, cur_pos=0):
        """
        rewind the buffer
        """
        self._cur_pos = cur_pos
        self._size = len(self._data) - cur_pos


def create_nonce(size=32):
    return os.urandom(size)



import pytest
from itertools import chain

from asyncua import ua, Client
from asyncua.client.ua_client import UASocketProtocol
from asyncua.common.subscription import Subscription
from asyncua.crypto import security_policies
from asyncua.ua.uaerrors import BadSessionClosed
from asyncua.client.ha.ha_client import (
    ConnectionStates,
    HaClient,
    HaMode,
    HaSecurityConfig,
)
from asyncua.client.ha.common import ClientNotFound
from asyncua.crypto.uacrypto import CertProperties
from .test_subscriptions import MySubHandler
from .conftest import (
    wait_clients_socket,
    wait_for_status_change,
    wait_mode_in_real_map,
    wait_node_in_real_map,
    wait_sub_in_real_map,
)


class TestHaClient:
    @pytest.mark.asyncio
    async def test_init_ha_client(self, ha_client):
        # check init parameters are set
        ha_config = ha_client._config
        urls = set(ha_client.urls)
        assert not ha_client.sub_names
        assert not ha_client.url_to_reset
        assert next(ha_client._gen_sub) == "sub_0"
        assert len(ha_client.get_clients()) == 2
        # check lower level clients have their parameters passed
        for client in ha_client.get_clients():
            url = client.server_url.geturl()
            assert url in urls
            urls.remove(url)
            assert client.session_timeout == ha_config.session_timeout * 1000
            assert client.session_timeout == ha_client.session_timeout * 1000
            assert client.secure_channel_timeout == ha_config.secure_channel_timeout * 1000
            assert client.uaclient._timeout == ha_config.request_timeout
            assert client.description == ha_config.session_name
            # ideal map is empty at startup
            assert not ha_client.ideal_map[url]
        assert not urls

    @pytest.mark.asyncio
    async def test_clients_connected(self, ha_client):
        # check lower level clients are connected
        await ha_client.start()
        await wait_clients_socket(ha_client, UASocketProtocol.OPEN)
        for client, srv_info in ha_client.clients.items():
            # TODO: Enable this when freeopcua server fully supports Service Levels
            # assert srv_info.status == 255
            assert srv_info.status
            assert srv_info.url == client.server_url.geturl()

        await ha_client.stop()
        await wait_clients_socket(ha_client, UASocketProtocol.CLOSED)

    @pytest.mark.asyncio
    async def test_all_tasks_running(self, ha_client):
        # check all tasks are running
        await ha_client.start()
        tasks = list(
            chain(
                ha_client._manager_task.values(),
                ha_client._keepalive_task.values(),
                ha_client._manager_task.values(),
                ha_client._reconciliator_task.values(),
            )
        )
        task_objs = list(
            chain(
                ha_client._manager_task,
                ha_client._keepalive_task,
                ha_client._manager_task,
                ha_client._reconciliator_task,
            )
        )
        task_count = 0
        for obj in task_objs:
            task_count += 1
            assert not obj.stop_event.is_set()
        assert task_count == 5

        for task in tasks:
            assert not task.done()
            assert not task.cancelled()

        await ha_client.stop()

        for obj in task_objs:
            assert obj.stop_event.is_set()

        for task in tasks:
            assert task.done()

    @pytest.mark.asyncio
    async def test_subscription(self, ha_client, srv_variables):

        await ha_client.start()
        node, values = [(n, v) for n, v in srv_variables.items()][0]

        myhandler = MySubHandler()
        sub = await ha_client.create_subscription(100, myhandler)
        await ha_client.subscribe_data_change(sub, [node])

        # the subscription is immediately available in ideal_map
        for c in ha_client.get_clients():
            url = c.server_url.geturl()
            assert sub in ha_client.ideal_map[url]

        # keepalive (status change) and reconciliator (real_mode_for_sub)
        # must both run once to assess the outcome.
        await wait_sub_in_real_map(ha_client, sub)
        for c in ha_client.get_clients():
            await wait_for_status_change(ha_client, c, 255)

        primary = await ha_client.get_serving_client(ha_client.get_clients(), None)
        await wait_mode_in_real_map(
            ha_client, primary, sub, mode="publishing", value=True
        )

        node_id, val, data = await myhandler.result()
        assert node_id == node
        assert val == values

        # real map data are OK
        monitoring = {ua.MonitoringMode.Reporting, ua.MonitoringMode.Disabled}
        publishing = {True, False}
        node_str = None

        for vsub in ha_client.reconciliator.real_map.values():
            vs = vsub[sub]
            node_str = node.nodeid.to_string()
            assert next(iter(vs.nodes)) == node_str
            assert vs.nodes[node_str].queuesize == 0
            assert vs.nodes[node_str].attr == ua.AttributeIds.Value
            assert vs.period == 100
            assert vs.handler == myhandler
            assert vs.monitoring in monitoring
            monitoring.remove(vs.monitoring)
            assert vs.publishing in publishing
            publishing.remove(vs.publishing)

        # in warm mode: 1 client publishes / 1 client monitor
        assert not publishing
        assert not monitoring

        # unsubscribe remove nodes from real map
        await ha_client.unsubscribe([node])
        await wait_node_in_real_map(ha_client, sub, node_str, negation=True)

        # delete subscription removes vs from real map
        await ha_client.delete_subscriptions([sub])
        await wait_sub_in_real_map(ha_client, sub, negation=True)

    @pytest.mark.asyncio
    async def test_reconnect(self, ha_client):
        await ha_client.start()
        await wait_clients_socket(ha_client, UASocketProtocol.OPEN)

        first_client = next(iter(ha_client.get_clients()))
        socket = first_client.uaclient.protocol
        await ha_client.reconnect(first_client)

        assert first_client.server_url.geturl() in ha_client.url_to_reset
        await wait_clients_socket(ha_client, UASocketProtocol.OPEN)
        new_socket = first_client.uaclient.protocol
        assert socket != new_socket

        # urls to reconnect should only be added once
        await ha_client.reconciliator.stop()
        await ha_client.reconnect(first_client)
        await ha_client.reconnect(first_client)
        assert len(ha_client.url_to_reset) == 2

    @pytest.mark.asyncio
    async def test_failover_warm(self, ha_client, srv_variables):
        await ha_client.start()
        node, values = [(n, v) for n, v in srv_variables.items()][0]

        myhandler = MySubHandler()
        sub = await ha_client.create_subscription(100, myhandler)
        await ha_client.subscribe_data_change(sub, [node])

        # data collected by sub handler is correct
        await wait_sub_in_real_map(ha_client, sub)
        node_id, val, data = await myhandler.result()
        assert node_id == node
        assert val == values

        # real map data are OK
        reconciliator = ha_client.reconciliator
        _url = None
        for _url, vsub in reconciliator.real_map.items():
            vs = vsub[sub]
            if vs.publishing:
                break

        primary_client = ha_client.get_client_by_url(_url)
        secondary_client = set(ha_client.get_clients()) - {primary_client}
        await ha_client.failover_warm(secondary_client.pop(), {primary_client})

        # hack to wait for the next reconciliator iteration
        sub2 = await ha_client.create_subscription(100, myhandler)
        await wait_sub_in_real_map(ha_client, sub2)

        for _url, vsub in reconciliator.real_map.items():
            vs = vsub[sub]
            if vs.publishing:
                break

        new_primary = ha_client.get_client_by_url(_url)
        assert primary_client != new_primary

    @pytest.mark.asyncio
    async def test_security(self, ha_config, ha_servers, client_key_and_cert):
        # check security can be set via method and constructor
        ha_clients = []
        srv1, srv2 = ha_servers
        srv1_port = srv1.endpoint.port
        srv2_port = srv2.endpoint.port

        ha_config = ha_config(srv1_port, srv2_port)
        ha_c1 = HaClient(ha_config)
        key, cert = client_key_and_cert

        security_policy = security_policies.SecurityPolicyBasic256Sha256
        user_key = CertProperties(key, "PEM")
        user_cert = CertProperties(cert, "DER")
        security_mode = ua.MessageSecurityMode.SignAndEncrypt
        ha_c1.set_security(security_policy, user_cert, user_key, mode=security_mode)
        ha_clients.append(ha_c1)

        ha_security = HaSecurityConfig(
            security_policy, user_cert, user_key, mode=security_mode
        )
        ha_clients.append(HaClient(ha_config, ha_security))

        for ha_client in ha_clients:
            assert ha_client.security_config.policy == security_policy
            assert ha_client.security_config.mode == security_mode
            assert ha_client.security_config.certificate == user_cert
            assert ha_client.security_config.private_key == user_key

            await ha_client.start()
            await wait_clients_socket(ha_client, UASocketProtocol.OPEN)

            for client in ha_client.get_clients():
                assert isinstance(client.security_policy, ua.SecurityPolicy)
                assert isinstance(client.uaclient.security_policy, ua.SecurityPolicy)
                assert client.security_policy.Mode == security_mode
                assert client.security_policy.peer_certificate

                # security policy should be cleared once reconnect is called
                policy = client.security_policy
                await ha_client.reconnect(client)
                assert client.security_policy != policy
                assert client.security_policy.Mode == security_mode
                assert client.security_policy.peer_certificate
            await ha_client.stop()

    @pytest.mark.asyncio
    async def test_group_clients_by_health(self, ha_client, ha_servers):
        srv1, srv2 = ha_servers
        # srv2 service level is already 255
        slevel = srv1.get_node(ua.NodeId(ua.ObjectIds.Server_ServiceLevel))
        await ha_client.start()
        for c in ha_client.get_clients():
            await wait_for_status_change(ha_client, c, 255)

        # if all clients are 255, group client should return them all
        healthy, unhealthy = await ha_client.group_clients_by_health()
        assert len(healthy) == 2
        assert not unhealthy

        # the service level is considered (by default) unhealthy below 200,
        # so change to unhealthy for srv1 and client1
        clients = ha_client.get_clients()
        await slevel.write_value(ua.Variant(199, ua.VariantType.Byte))
        await wait_for_status_change(ha_client, clients[0], 199)
        healthy, unhealthy = await ha_client.group_clients_by_health()
        assert clients[0] == unhealthy[0]
        assert clients[1] == healthy[0]

        # now try with a custom HEALTHY_STATE value
        ha_client.HEALTHY_STATE = 4
        healthy, unhealthy = await ha_client.group_clients_by_health()
        assert len(healthy) == 2
        assert not unhealthy

        await slevel.write_value(ua.Variant(3, ua.VariantType.Byte))
        await wait_for_status_change(ha_client, clients[0], 3)
        healthy, unhealthy = await ha_client.group_clients_by_health()
        assert clients[0] == unhealthy[0]
        assert clients[1] == healthy[0]

        # set back the value to 255 since the fixture as a wide scope
        await slevel.write_value(ua.Variant(255, ua.VariantType.Byte))

    @pytest.mark.asyncio
    async def test_ha_mode(self, ha_config, ha_servers):
        # WARM is the only mode supported at the moment.
        srv1, srv2 = ha_servers
        port1, port2 = srv1.endpoint.port, srv2.endpoint.port
        config = ha_config(port1, port2, ha_mode=HaMode.WARM)
        ha_client = HaClient(config)
        assert ha_client.ha_mode == "warm"

        for ha in (HaMode.COLD, HaMode.HOT_A, HaMode.HOT_B):
            config = ha_config(port1, port2, ha_mode=ha)
            with pytest.raises(NotImplementedError):
                ha_client = HaClient(config)

    @pytest.mark.asyncio
    async def test_get_client_by_url(self, ha_client):
        urls = [srv_info.url for srv_info in ha_client.clients.values()]
        for url in urls:
            assert isinstance(ha_client.get_client_by_url(url), Client)
        with pytest.raises(ClientNotFound):
            ha_client.get_client_by_url("opc.tcp://not_found")


class TestKeepAlive:
    @pytest.mark.asyncio
    @pytest.mark.parametrize("excp", [Exception(), BadSessionClosed()])
    async def test_service_level_no_data(self, excp, ha_client, mocker):
        mock_read = mocker.patch.object(Client, "read_values")
        mock_read.return_value = excp
        await ha_client.start()
        for client in ha_client.get_clients():
            await wait_for_status_change(ha_client, client, ConnectionStates.NO_DATA)


class TestHaManager:
    @pytest.mark.asyncio
    async def test_reconnect(self, ha_client):
        await ha_client.start()
        await wait_clients_socket(ha_client, UASocketProtocol.OPEN)
        for c in ha_client.get_clients():
            await c.disconnect()
        # the HaManager task should reconnect all the clients automagically
        await wait_clients_socket(ha_client, UASocketProtocol.OPEN)


class TestReconciliator:
    @pytest.mark.asyncio
    async def test_resubscribe(self, ha_client, srv_variables):
        """
        Here we test that when the HaClient reconnect a client,
        Reconciliator drop any existing subscription and create new one.
        """
        await ha_client.start()
        for c in ha_client.get_clients():
            await wait_for_status_change(ha_client, c, 255)

        node, values = [(n, v) for n, v in srv_variables.items()][0]
        node_str = node.nodeid.to_string()

        # create sub and add node to datachange
        myhandler = MySubHandler()
        sub = await ha_client.create_subscription(100, myhandler)
        await ha_client.subscribe_data_change(sub, [node])
        # wait for the node to appear in real map
        await wait_node_in_real_map(ha_client, sub, node_str)

        # the reconciliator subscription and handle trackers must be up-to-date.
        reconciliator = ha_client.reconciliator
        clients = ha_client.get_clients()
        for client in clients:
            url = client.server_url.geturl()
            assert sub in reconciliator.name_to_subscription[url]
            assert isinstance(
                reconciliator.name_to_subscription[url][sub], Subscription
            )
            assert isinstance(reconciliator.node_to_handle[url][node_str], int)

        primary = await ha_client.get_serving_client(ha_client.get_clients(), None)
        # let's save the primary sub/handle for future comparison
        primary_url = primary.server_url.geturl()
        real_sub = reconciliator.name_to_subscription[primary_url][sub]
        real_handle = reconciliator.node_to_handle[primary_url][node_str]

        # Add to url_to_reset triggers reconciliator to resub
        async with ha_client._url_to_reset_lock:
            ha_client.url_to_reset.add(primary_url)

        # hack to wait for the reconciliator iteration
        sub2 = await ha_client.create_subscription(100, myhandler)
        await wait_sub_in_real_map(ha_client, sub2)

        new_real_handle = reconciliator.node_to_handle[primary_url][node_str]
        new_real_sub = reconciliator.name_to_subscription[primary_url][sub]
        # Once resub, the low-level client subscription must be brand new.
        assert real_sub != new_real_sub
        # However we haven't deleted the MI, so the server returns the same handle
        assert real_handle == new_real_handle

    @pytest.mark.asyncio
    async def test_subscribe_mi_with_batch(self, ha_client, srv_variables, mocker):
        """
        Ensure we are not exceding the threshold of MonitoredItems
        per request when subscribing/unsubscribing.
        """
        await ha_client.start()

        reconciliator = ha_client.reconciliator
        reconciliator.BATCH_MI_SIZE = 1
        myhandler = MySubHandler()
        sub = await ha_client.create_subscription(100, myhandler)
        await wait_sub_in_real_map(ha_client, sub)

        # once the real subscription is created patch it for one client
        first_client = next(iter(ha_client.get_clients()))
        url = first_client.server_url.geturl()
        real_sub = reconciliator.name_to_subscription[url][sub]

        mock_subscribe_data_change = mocker.patch.object(
            real_sub, "subscribe_data_change", wraps=real_sub.subscribe_data_change
        )
        mock_unsubscribe = mocker.patch.object(
            real_sub, "unsubscribe", wraps=real_sub.unsubscribe
        )

        node_list = [n.nodeid.to_string() for n in srv_variables]

        # wait for the node to appear in real map
        await ha_client.subscribe_data_change(sub, node_list)
        for n in node_list:
            await wait_node_in_real_map(ha_client, sub, n)
        assert mock_subscribe_data_change.call_count == 2

        await ha_client.unsubscribe(node_list)
        for n in node_list:
            await wait_node_in_real_map(ha_client, sub, n, negation=True)
        assert mock_unsubscribe.call_count == 2

        reconciliator.BATCH_MI_SIZE = 2
        mock_subscribe_data_change.reset_mock()
        mock_unsubscribe.reset_mock()

        await ha_client.subscribe_data_change(sub, node_list)
        # wait for the node to appear in real map
        for n in node_list:
            await wait_node_in_real_map(ha_client, sub, n)
        assert mock_subscribe_data_change.call_count == 1

        await ha_client.unsubscribe(node_list)
        for n in node_list:
            await wait_node_in_real_map(ha_client, sub, n, negation=True)
        assert mock_unsubscribe.call_count == 1

    @pytest.mark.asyncio
    async def test_remove_bad_nodes_from_both_maps(
        self, ha_client, srv_variables, mocker
    ):
        """
        When a server respond to a MI request with BadUnknownId, make sure
        we remove the faulty node from the ideal and the real map. Otherwhise,
        the ideal_map is bad, and Reconciliator would try to comply indefinitely.
        """
        await ha_client.start()
        node, values = [(n, v) for n, v in srv_variables.items()][0]
        node_str = node.nodeid.to_string()
        myhandler = MySubHandler()
        sub = await ha_client.create_subscription(100, myhandler)
        node_fake = f"{node_str}0"
        node_list = [node_str, node_fake]
        await ha_client.subscribe_data_change(sub, node_list)
        await wait_node_in_real_map(ha_client, sub, node_str)
        reconciliator = ha_client.reconciliator

        for c in ha_client.get_clients():
            url = c.server_url.geturl()
            assert node_str in reconciliator.real_map[url][sub].nodes
            assert node_str in ha_client.ideal_map[url][sub].nodes
            assert node_fake not in reconciliator.real_map[url][sub].nodes
            assert node_fake not in ha_client.ideal_map[url][sub].nodes



import logging
from enum import Enum
from typing import Coroutine, Iterable, Optional

from asyncua import ua
from ..common.callback import CallbackType, ServerItemCallback
from ..common.utils import create_nonce, ServiceError
from .address_space import AddressSpace
from .users import User, UserRole
from .subscription_service import SubscriptionService


class SessionState(Enum):
    Created = 0
    Activated = 1
    Closed = 2


class InternalSession:
    """

    """
    max_connections = 1000
    _current_connections = 0
    _counter = 10
    _auth_counter = 1000

    def __init__(self, internal_server, aspace: AddressSpace, submgr: SubscriptionService, name,
                 user=User(role=UserRole.Anonymous), external=False):
        self.logger = logging.getLogger(__name__)
        self.iserver = internal_server
        # define if session is external, we need to copy some objects if it is internal
        self.external = external
        self.aspace: AddressSpace = aspace
        self.subscription_service: SubscriptionService = submgr
        self.name = name
        self.user = user
        self.nonce = None
        self.state = SessionState.Created
        self.session_id = ua.NodeId(self._counter)
        InternalSession._counter += 1
        self.subscriptions = []
        self.auth_token = ua.NodeId(self._auth_counter)
        InternalSession._auth_counter += 1
        self.logger.info('Created internal session %s', self.name)

    def __str__(self):
        return f'InternalSession(name:{self.name},' \
               f' user:{self.user}, id:{self.session_id}, auth_token:{self.auth_token})'

    async def get_endpoints(self, params=None, sockname=None):
        return await self.iserver.get_endpoints(params, sockname)

    async def create_session(self, params, sockname=None):
        self.logger.info('Create session request')
        result = ua.CreateSessionResult()
        result.SessionId = self.session_id
        result.AuthenticationToken = self.auth_token
        result.RevisedSessionTimeout = params.RequestedSessionTimeout
        result.MaxRequestMessageSize = 65536
        self.nonce = create_nonce(32)
        result.ServerNonce = self.nonce
        result.ServerEndpoints = await self.get_endpoints(sockname=sockname)

        return result

    async def close_session(self, delete_subs=True):
        self.logger.info('close session %s', self.name)
        if self.state == SessionState.Activated:
            InternalSession._current_connections -= 1
        if InternalSession._current_connections < 0:
            InternalSession._current_connections = 0
        self.state = SessionState.Closed
        await self.delete_subscriptions(self.subscriptions)

    def activate_session(self, params, peer_certificate):
        self.logger.info('activate session')
        result = ua.ActivateSessionResult()
        if self.state != SessionState.Created:
            raise ServiceError(ua.StatusCodes.BadSessionIdInvalid)
        if InternalSession._current_connections >= InternalSession.max_connections:
            raise ServiceError(ua.StatusCodes.BadMaxConnectionsReached)
        self.nonce = create_nonce(32)
        result.ServerNonce = self.nonce
        for _ in params.ClientSoftwareCertificates:
            result.Results.append(ua.StatusCode())
        self.state = SessionState.Activated
        InternalSession._current_connections += 1
        id_token = params.UserIdentityToken
        if self.iserver.user_manager is not None:
            if isinstance(id_token, ua.UserNameIdentityToken):
                username, password = self.iserver.check_user_token(self, id_token)
            else:
                username, password = None, None

            user = self.iserver.user_manager.get_user(self.iserver, username=username, password=password,
                                                      certificate=peer_certificate)
            if user is None:
                raise ServiceError(ua.StatusCodes.BadUserAccessDenied)
            else:
                self.user = user
        self.logger.info("Activated internal session %s for user %s", self.name, self.user)
        return result

    async def read(self, params):
        if self.user is None:
            user = User()
        else:
            user = self.user
        await self.iserver.callback_service.dispatch(CallbackType.PreRead,
                                                     ServerItemCallback(params, None, user, self.external))
        results = self.iserver.attribute_service.read(params)
        await self.iserver.callback_service.dispatch(CallbackType.PostRead,
                                                     ServerItemCallback(params, results, user, self.external))
        return results

    async def history_read(self, params) -> Coroutine:
        return await self.iserver.history_manager.read_history(params)

    async def write(self, params):
        if self.user is None:
            user = User()
        else:
            user = self.user
        await self.iserver.callback_service.dispatch(CallbackType.PreWrite,
                                                     ServerItemCallback(params, None, user, self.external))
        write_result = await self.iserver.attribute_service.write(params, user=user)
        await self.iserver.callback_service.dispatch(CallbackType.PostWrite,
                                                     ServerItemCallback(params, write_result, user, self.external))
        return write_result

    async def browse(self, params):
        return self.iserver.view_service.browse(params)

    async def translate_browsepaths_to_nodeids(self, params):
        return self.iserver.view_service.translate_browsepaths_to_nodeids(params)

    async def add_nodes(self, params):
        return self.iserver.node_mgt_service.add_nodes(params, self.user)

    async def delete_nodes(self, params):
        return self.iserver.node_mgt_service.delete_nodes(params, self.user)

    async def add_references(self, params):
        return self.iserver.node_mgt_service.add_references(params, self.user)

    async def delete_references(self, params):
        return self.iserver.node_mgt_service.delete_references(params, self.user)

    def add_method_callback(self, methodid, callback):
        return self.aspace.add_method_callback(methodid, callback)

    async def call(self, params):
        """COROUTINE"""
        return await self.iserver.method_service.call(params)

    async def create_subscription(self, params, callback=None):
        result = await self.subscription_service.create_subscription(params, callback, external=self.external)
        self.subscriptions.append(result.SubscriptionId)
        return result

    async def create_monitored_items(self, params: ua.CreateMonitoredItemsParameters):
        """Returns Future"""
        subscription_result = await self.subscription_service.create_monitored_items(params)
        await self.iserver.callback_service.dispatch(CallbackType.ItemSubscriptionCreated,
                                                     ServerItemCallback(params, subscription_result, None,
                                                                        self.external))
        return subscription_result

    async def modify_monitored_items(self, params):
        subscription_result = self.subscription_service.modify_monitored_items(params)
        await self.iserver.callback_service.dispatch(CallbackType.ItemSubscriptionModified,
                                                     ServerItemCallback(params, subscription_result, None,
                                                                        self.external))
        return subscription_result

    def republish(self, params):
        return self.subscription_service.republish(params)

    async def delete_subscriptions(self, ids):
        # This is an async method, dues to symmetry with client code
        return await self.subscription_service.delete_subscriptions(ids)

    async def delete_monitored_items(self, params):
        # This is an async method, dues to symmetry with client code
        subscription_result = self.subscription_service.delete_monitored_items(params)
        await self.iserver.callback_service.dispatch(CallbackType.ItemSubscriptionDeleted,
                                                     ServerItemCallback(params, subscription_result, None,
                                                                        self.external))

        return subscription_result

    def publish(self, acks: Optional[Iterable[ua.SubscriptionAcknowledgement]] = None):
        return self.subscription_service.publish(acks or [])

    def modify_subscription(self, params, callback):
        return self.subscription_service.modify_subscription(params, callback)



from asyncua import ua


class UaFile:

    def __init__(self, file_node, open_mode):
        self._file_node = file_node
        self._handle = None
        if open_mode == 'r':
            self._init_open = ua.OpenFileMode.Read.value
        else:
            raise ValueError("file mode is not supported")

    async def __aenter__(self):
        self._handle = await self.open(self._init_open)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        return await self.close()

    async def open(self, open_mode):
        """ open file method """
        open_node = await self._file_node.get_child("Open")
        arg = ua.Variant(open_mode, ua.VariantType.Byte)
        return await self._file_node.call_method(open_node, arg)

    async def close(self):
        """ close file method """
        read_node = await self._file_node.get_child("Close")
        arg1 = ua.Variant(self._handle, ua.VariantType.UInt32)
        return await self._file_node.call_method(read_node, arg1)

    async def read(self):
        """ reads file contents """
        size = await self.get_size()
        read_node = await self._file_node.get_child("Read")
        arg1 = ua.Variant(self._handle, ua.VariantType.UInt32)
        arg2 = ua.Variant(size, ua.VariantType.Int32)
        return await self._file_node.call_method(read_node, arg1, arg2)

    async def get_size(self):
        """ gets size of file """
        size_node = await self._file_node.get_child("Size")
        return await size_node.read_value()




# the order is important, some classes are overriden
from .attribute_ids import AttributeIds
from .object_ids import ObjectIds
from .object_ids import ObjectIdNames
from .status_codes import StatusCodes
from .uaprotocol_auto import *
from .uaprotocol_hand import *
from .uatypes import *  # TODO: This should be renamed to uatypes_hand



import pytest
import os.path
import xml.etree.ElementTree as ET

from asyncua.server.address_space import AddressSpace
from asyncua.server.address_space import NodeManagementService
from asyncua.server.standard_address_space import standard_address_space


def find_elem(parent, name, ns=None):
    if ns is None:
        try:
            return parent.find(parent.tag[0:parent.tag.index('}') + 1] + name)
        except ValueError:
            return parent.find(name)
    return parent.find(ns + name)


def remove_elem(parent, name):
    e = find_elem(parent, name)
    if e is not None:
        parent.remove(e)


def try_apply(item, aliases):
    attrib = item.attrib
    for name in ('ReferenceType', 'DataType'):
        try:
            value = attrib[name]
            attrib[name] = aliases[value]
        except KeyError:
            pass


def read_nodes(path):
    tree = ET.parse(path)
    root = tree.getroot()
    aliases_elem = find_elem(root, 'Aliases')
    aliases = dict((a.attrib['Alias'], a.text) for a in aliases_elem)
    any(try_apply(i, aliases) for i in root.iter())
    root.remove(aliases_elem)
    remove_elem(root, "Models")
    remove_elem(root, "NamespaceUris")
    return dict((e.attrib['NodeId'], e) for e in root)


def get_refs(e):
    return set((r.attrib['ReferenceType'], r.text, r.attrib.get('IsForward', 'true') == 'true') for r in
               find_elem(e, 'References'))


@pytest.mark.skip("Donot understand that code and I am not sure we should test the xml file. it is not ours")
def test_std_address_space_references():
    aspace = AddressSpace()
    node_mgt_service = NodeManagementService(aspace)
    standard_address_space.fill_address_space(node_mgt_service)
    std_nodes = read_nodes(
        os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'schemas', 'Opc.Ua.NodeSet2.xml'))
    )
    for k in aspace.keys():
        refs = set(
            (r.ReferenceTypeId.to_string(), r.NodeId.to_string(), r.IsForward) for r in aspace[k].references
        )
        xml_refs = set(
            (r.attrib['ReferenceType'], r.text, r.attrib.get('IsForward', 'true') == 'true') for r in
                       find_elem(std_nodes[k.to_string()], 'References')
        )
        assert 0 == len(xml_refs - refs)



"""
Support for custom structures in client and server
We only support a subset of features but should be enough
for custom structures
"""

import uuid
import logging
# The next two imports are for generated code
from datetime import datetime
from enum import IntEnum, EnumMeta
from dataclasses import dataclass, field
from typing import List


from xml.etree import ElementTree as ET
from asyncua import ua

from .structures104 import get_default_value, clean_name

_logger = logging.getLogger(__name__)


class EnumType(object):
    def __init__(self, name):
        self.name = clean_name(name)
        self.fields = []
        self.typeid = None

    def __str__(self):
        return f"EnumType({self.name, self.fields})"

    __repr__ = __str__

    def get_code(self):
        code = """

class {0}(IntEnum):

    '''
    {0} EnumInt autogenerated from xml
    '''

""".format(self.name)

        for EnumeratedValue in self.fields:
            name = clean_name(EnumeratedValue.Name)
            value = EnumeratedValue.Value
            code += f"    {name} = {value}\n"

        return code


class EnumeratedValue:
    def __init__(self, name, value):
        if name == "None":
            name = "None_"
        name = name.replace(" ", "")
        self.Name = name
        self.Value = value


class Struct:
    def __init__(self, name):
        self.name = clean_name(name)
        self.fields = []
        self.typeid = None

    def __str__(self):
        return f"Struct(name={self.name}, fields={self.fields}"

    __repr__ = __str__

    def get_code(self):
        code = f"""

@dataclass
class {self.name}:

    '''
    {self.name} structure autogenerated from xml
    '''

"""
        for sfield in self.fields:
            uatype = f"ua.{sfield.uatype}"
            if sfield.array:
                uatype = f"List[{uatype}]"
            if uatype == 'List[ua.Char]':
                uatype = 'String'
            code += f"    {sfield.name}:{uatype} = {sfield.value}\n"
        print(code)
        return code


class Field(object):
    def __init__(self, name):
        self.name = name
        self.uatype = None
        self.value = None
        self.array = False

    def __str__(self):
        return f"Field(name={self.name}, uatype={self.uatype})"

    __repr__ = __str__


class StructGenerator(object):
    def __init__(self):
        self.model = []

    def make_model_from_string(self, xml):
        obj = ET.fromstring(xml)
        self._make_model(obj)

    def make_model_from_file(self, path):
        obj = ET.parse(path)
        root = obj.getroot()
        self._make_model(root)

    def _make_model(self, root):
        enums = {}
        for child in root:
            if child.tag.endswith("EnumeratedType"):
                intenum = EnumType(child.get("Name"))
                for xmlfield in child:
                    if xmlfield.tag.endswith("EnumeratedValue"):
                        name = xmlfield.get("Name")
                        value = xmlfield.get("Value")
                        enumvalue = EnumeratedValue(name, value)
                        intenum.fields.append(enumvalue)
                        enums[child.get("Name")] = value
                self.model.append(intenum)

        for child in root:
            if child.tag.endswith("StructuredType"):
                struct = Struct(child.get("Name"))
                array = False
                # these lines can be reduced in >= Python3.8 with root.iterfind("{*}Field") and similar
                for xmlfield in child:
                    if xmlfield.tag.endswith("Field"):
                        name = xmlfield.get("Name")
                        if name.startswith("NoOf"):
                            array = True
                            continue
                        field = Field(clean_name(name))
                        field.uatype = xmlfield.get("TypeName")
                        if ":" in field.uatype:
                            field.uatype = field.uatype.split(":")[1]
                        field.uatype = clean_name(field.uatype)
                        field.value = get_default_value(field.uatype, enums)
                        if array:
                            field.array = True
                            field.value = "field(default_factory=list)"
                            array = False
                        struct.fields.append(field)
                self.model.append(struct)

    def save_to_file(self, path, register=False):
        _file = open(path, "w+")
        self._make_header(_file)
        for struct in self.model:
            _file.write(struct.get_code())
        if register:
            _file.write(self._make_registration())
        _file.close()

    def _make_registration(self):
        code = "\n\n"
        for struct in self.model:
            code += f"ua.register_extension_object('{struct.name}'," \
                    f" ua.NodeId.from_string('{struct.typeid}'), {struct.name})\n"
        return code

    def get_python_classes(self, env=None):
        return _generate_python_class(self.model, env=env)

    def _make_header(self, _file):
        _file.write("""
'''
THIS FILE IS AUTOGENERATED, DO NOT EDIT!!!
'''

from datetime import datetime
import uuid
from dataclasses import dataclass, field
from typing import List, Union

from asyncua import ua
""")

    def set_typeid(self, name, typeid):
        for struct in self.model:
            if struct.name == name:
                struct.typeid = typeid
                return


async def load_type_definitions(server, nodes=None):
    """
    Download xml from given variable node defining custom structures.
    If no node is given, attemps to import variables from all nodes under
    "0:OPC Binary"
    the code is generated and imported on the fly. If you know the structures
    are not going to be modified it might be interresting to copy the generated files
    and include them in you code
    """
    if nodes is None:
        nodes = []
        for desc in await server.nodes.opc_binary.get_children_descriptions():
            if desc.BrowseName != ua.QualifiedName("Opc.Ua"):
                nodes.append(server.get_node(desc.NodeId))

    structs_dict = {}
    generators = []
    for node in nodes:
        xml = await node.read_value()
        generator = StructGenerator()
        generators.append(generator)
        generator.make_model_from_string(xml)
        # generate and execute new code on the fly
        generator.get_python_classes(structs_dict)
        # same but using a file that is imported. This can be useful for debugging library
        # name = node.read_browse_name().Name
        # Make sure structure names do not contain charaters that cannot be used in Python class file names
        # name = clean_name(name)
        # name = "structures_" + node.read_browse_name().Name
        # generator.save_and_import(name + ".py", append_to=structs_dict)

        # register classes
        # every children of our node should represent a class
        for ndesc in await node.get_children_descriptions():
            ndesc_node = server.get_node(ndesc.NodeId)
            ref_desc_list = await ndesc_node.get_references(refs=ua.ObjectIds.HasDescription, direction=ua.BrowseDirection.Inverse)
            if ref_desc_list:  # some server put extra things here
                name = clean_name(ndesc.BrowseName.Name)
                if name not in structs_dict:
                    _logger.warning("%s is found as child of binary definition node but is not found in xml", name)
                    continue
                nodeid = ref_desc_list[0].NodeId
                ua.register_extension_object(name, nodeid, structs_dict[name])
                # save the typeid if user want to create static file for type definitnion
                generator.set_typeid(name, nodeid.to_string())

        for key, val in structs_dict.items():
            if isinstance(val, EnumMeta) and key != "IntEnum":
                setattr(ua, key, val)

    return generators, structs_dict


def _generate_python_class(model, env=None):
    """
    generate Python code and execute in a new environment
    return a dict of structures {name: class}
    Rmw: Since the code is generated on the fly, in case of error the stack trace is
    not available and debugging is very hard...
    """
    if env is None:
        env = ua.__dict__
    #  Add the required libraries to dict
    if "ua" not in env:
        env['ua'] = ua
    if "datetime" not in env:
        env['datetime'] = datetime
    if "uuid" not in env:
        env['uuid'] = uuid
    if "enum" not in env:
        env['IntEnum'] = IntEnum
    if "dataclass" not in env:
        env['dataclass'] = dataclass
    if "field" not in env:
        env['field'] = field
    if "List" not in env:
        env['List'] = List
    # generate classes one by one and add them to dict
    for element in model:
        code = element.get_code()
        try:
            exec(code, env)
        except Exception:
            _logger.exception("Failed to execute auto-generated code from UA datatype: %s", code)
            raise
    return env


async def load_enums(server, env=None, force=False):
    """
    Read enumeration data types on server and generate python Enums in ua scope for them
    """
    model = []

    for desc in await server.nodes.enum_data_type.get_children_descriptions(refs=ua.ObjectIds.HasSubtype):
        enum_name = desc.BrowseName.Name
        enum_node = server.get_node(desc.NodeId)
        if not force and hasattr(ua, enum_name):
            _logger.debug("Enum type %s is already in ua namespace, ignoring", enum_name)
            continue
        c = None
        for child_desc in await enum_node.get_children_descriptions(refs=ua.ObjectIds.HasProperty):
            child_node = server.get_node(child_desc.NodeId)
            if child_desc.BrowseName.Name == "EnumStrings":
                c = await _get_enum_strings(enum_name, child_node)
            elif child_desc.BrowseName.Name == "EnumValues":
                c = await _get_enum_values(enum_name, server.get_node(child_desc.NodeId))
            else:
                _logger.warning("Unexpected children of node %s: %s", desc, child_desc)
        if c is not None:
            model.append(c)
    return _generate_python_class(model, env=env)


async def _get_enum_values(name, node):
    val = await node.read_value()
    c = EnumType(name)
    c.fields = [EnumeratedValue(enumval.DisplayName.Text, enumval.Value) for enumval in val]
    return c


async def _get_enum_strings(name, node):
    val = await node.read_value()
    c = EnumType(name)
    c.fields = [EnumeratedValue(st.Text, idx) for idx, st in enumerate(val)]
    return c




"""
Autogenerated code from xml"

Model Uri:http://opcfoundation.org/UA/"
Version:1.04.9"
Publication date:2021-01-21T00:00:00Z"

File creation Date:2021-04-06 17:23:55.234980"
"""
from asyncua import ua
from .events import Event


__all__ = ["BaseEvent", "AuditEvent", "AuditSecurityEvent", "AuditChannelEvent", "AuditOpenSecureChannelEvent", "AuditSessionEvent", "AuditCreateSessionEvent", "AuditActivateSessionEvent", "AuditCancelEvent", "AuditCertificateEvent", "AuditCertificateDataMismatchEvent", "AuditCertificateExpiredEvent", "AuditCertificateInvalidEvent", "AuditCertificateUntrustedEvent", "AuditCertificateRevokedEvent", "AuditCertificateMismatchEvent", "AuditNodeManagementEvent", "AuditAddNodesEvent", "AuditDeleteNodesEvent", "AuditAddReferencesEvent", "AuditDeleteReferencesEvent", "AuditUpdateEvent", "AuditWriteUpdateEvent", "AuditHistoryUpdateEvent", "AuditUpdateMethodEvent", "SystemEvent", "DeviceFailureEvent", "BaseModelChangeEvent", "GeneralModelChangeEvent", "TransitionEvent", "AuditUpdateStateEvent", "ProgramTransitionEvent", "SemanticChangeEvent", "AuditUrlMismatchEvent", "Condition", "RefreshStartEvent", "RefreshEndEvent", "RefreshRequiredEvent", "AuditConditionEvent", "AuditConditionEnableEvent", "AuditConditionCommentEvent", "DialogCondition", "AcknowledgeableCondition", "AlarmCondition", "LimitAlarm", "AuditHistoryEventUpdateEvent", "AuditHistoryValueUpdateEvent", "AuditHistoryDeleteEvent", "AuditHistoryRawModifyDeleteEvent", "AuditHistoryAtTimeDeleteEvent", "AuditHistoryEventDeleteEvent", "EventQueueOverflowEvent", "ProgramTransitionAuditEvent", "AuditConditionRespondEvent", "AuditConditionAcknowledgeEvent", "AuditConditionConfirmEvent", "ExclusiveLimitAlarm", "ExclusiveLevelAlarm", "ExclusiveRateOfChangeAlarm", "ExclusiveDeviationAlarm", "NonExclusiveLimitAlarm", "NonExclusiveLevelAlarm", "NonExclusiveRateOfChangeAlarm", "NonExclusiveDeviationAlarm", "DiscreteAlarm", "OffNormalAlarm", "TripAlarm", "AuditConditionShelvingEvent", "ProgressEvent", "SystemStatusChangeEvent", "SystemOffNormalAlarm", "AuditProgramTransitionEvent", "TrustListUpdatedAuditEvent", "CertificateUpdatedAuditEvent", "CertificateExpirationAlarm", "AuditConditionResetEvent", "PubSubStatusEvent", "PubSubTransportLimitsExceedEvent", "PubSubCommunicationFailureEvent", "DiscrepancyAlarm", "AuditConditionSuppressionEvent", "AuditConditionSilenceEvent", "AuditConditionOutOfServiceEvent", "RoleMappingRuleChangedAuditEvent", "KeyCredentialAuditEvent", "KeyCredentialUpdatedAuditEvent", "KeyCredentialDeletedAuditEvent", "InstrumentDiagnosticAlarm", "SystemDiagnosticAlarm", "AuditHistoryAnnotationUpdateEvent", "TrustListOutOfDateAlarm"]



class BaseEvent(Event):
    """
    BaseEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        Event.__init__(self)
        self.add_property('EventId', None, ua.VariantType.ByteString)
        self.add_property('EventType', ua.NodeId(ua.ObjectIds.BaseEventType), ua.VariantType.NodeId)
        self.add_property('SourceNode', sourcenode, ua.VariantType.NodeId)
        self.add_property('SourceName', None, ua.VariantType.String)
        self.add_property('Time', None, ua.VariantType.DateTime)
        self.add_property('ReceiveTime', None, ua.VariantType.DateTime)
        self.add_property('LocalTime', ua.NodeId(ua.ObjectIds.TimeZoneDataType), ua.VariantType.ExtensionObject)
        self.add_property('Message', ua.LocalizedText(message), ua.VariantType.LocalizedText)
        self.add_property('Severity', severity, ua.VariantType.UInt16)


class AuditEvent(BaseEvent):
    """
    AuditEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditEventType)
        self.add_property('ActionTimeStamp', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('Status', False, ua.VariantType.Boolean)
        self.add_property('ServerId', None, ua.VariantType.String)
        self.add_property('ClientAuditEntryId', None, ua.VariantType.String)
        self.add_property('ClientUserId', None, ua.VariantType.String)


class AuditSecurityEvent(AuditEvent):
    """
    AuditSecurityEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditSecurityEventType)
        self.add_property('StatusCodeId', None, ua.VariantType.StatusCode)


class AuditChannelEvent(AuditSecurityEvent):
    """
    AuditChannelEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditChannelEventType)
        self.add_property('SecureChannelId', None, ua.VariantType.String)


class AuditOpenSecureChannelEvent(AuditChannelEvent):
    """
    AuditOpenSecureChannelEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditOpenSecureChannelEventType)
        self.add_property('ClientCertificate', None, ua.VariantType.ByteString)
        self.add_property('ClientCertificateThumbprint', None, ua.VariantType.String)
        self.add_property('RequestType', None, ua.NodeId(ua.ObjectIds.SecurityTokenRequestType))
        self.add_property('SecurityPolicyUri', None, ua.VariantType.String)
        self.add_property('SecurityMode', None, ua.NodeId(ua.ObjectIds.MessageSecurityMode))
        self.add_property('RequestedLifetime', None, ua.NodeId(ua.ObjectIds.Duration))


class AuditSessionEvent(AuditSecurityEvent):
    """
    AuditSessionEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditSessionEventType)
        self.add_property('SessionId', ua.NodeId(ua.ObjectIds.AuditSessionEventType), ua.VariantType.NodeId)


class AuditCreateSessionEvent(AuditSessionEvent):
    """
    AuditCreateSessionEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCreateSessionEventType)
        self.add_property('SecureChannelId', None, ua.VariantType.String)
        self.add_property('ClientCertificate', None, ua.VariantType.ByteString)
        self.add_property('ClientCertificateThumbprint', None, ua.VariantType.String)
        self.add_property('RevisedSessionTimeout', None, ua.NodeId(ua.ObjectIds.Duration))


class AuditActivateSessionEvent(AuditSessionEvent):
    """
    AuditActivateSessionEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditActivateSessionEventType)
        self.add_property('ClientSoftwareCertificates', None, ua.NodeId(ua.ObjectIds.SignedSoftwareCertificate))
        self.add_property('UserIdentityToken', None, ua.NodeId(ua.ObjectIds.UserIdentityToken))
        self.add_property('SecureChannelId', None, ua.VariantType.String)


class AuditCancelEvent(AuditSessionEvent):
    """
    AuditCancelEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCancelEventType)
        self.add_property('RequestHandle', None, ua.VariantType.UInt32)


class AuditCertificateEvent(AuditSecurityEvent):
    """
    AuditCertificateEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCertificateEventType)
        self.add_property('Certificate', None, ua.VariantType.ByteString)


class AuditCertificateDataMismatchEvent(AuditCertificateEvent):
    """
    AuditCertificateDataMismatchEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCertificateDataMismatchEventType)
        self.add_property('InvalidHostname', None, ua.VariantType.String)
        self.add_property('InvalidUri', None, ua.VariantType.String)


class AuditCertificateExpiredEvent(AuditCertificateEvent):
    """
    AuditCertificateExpiredEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCertificateExpiredEventType)


class AuditCertificateInvalidEvent(AuditCertificateEvent):
    """
    AuditCertificateInvalidEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCertificateInvalidEventType)


class AuditCertificateUntrustedEvent(AuditCertificateEvent):
    """
    AuditCertificateUntrustedEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCertificateUntrustedEventType)


class AuditCertificateRevokedEvent(AuditCertificateEvent):
    """
    AuditCertificateRevokedEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCertificateRevokedEventType)


class AuditCertificateMismatchEvent(AuditCertificateEvent):
    """
    AuditCertificateMismatchEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditCertificateMismatchEventType)


class AuditNodeManagementEvent(AuditEvent):
    """
    AuditNodeManagementEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditNodeManagementEventType)


class AuditAddNodesEvent(AuditNodeManagementEvent):
    """
    AuditAddNodesEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditAddNodesEventType)
        self.add_property('NodesToAdd', None, ua.NodeId(ua.ObjectIds.AddNodesItem))


class AuditDeleteNodesEvent(AuditNodeManagementEvent):
    """
    AuditDeleteNodesEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditDeleteNodesEventType)
        self.add_property('NodesToDelete', None, ua.NodeId(ua.ObjectIds.DeleteNodesItem))


class AuditAddReferencesEvent(AuditNodeManagementEvent):
    """
    AuditAddReferencesEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditAddReferencesEventType)
        self.add_property('ReferencesToAdd', None, ua.NodeId(ua.ObjectIds.AddReferencesItem))


class AuditDeleteReferencesEvent(AuditNodeManagementEvent):
    """
    AuditDeleteReferencesEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditDeleteReferencesEventType)
        self.add_property('ReferencesToDelete', None, ua.NodeId(ua.ObjectIds.DeleteReferencesItem))


class AuditUpdateEvent(AuditEvent):
    """
    AuditUpdateEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditUpdateEventType)


class AuditWriteUpdateEvent(AuditUpdateEvent):
    """
    AuditWriteUpdateEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditWriteUpdateEventType)
        self.add_property('AttributeId', None, ua.VariantType.UInt32)
        self.add_property('IndexRange', None, ua.NodeId(ua.ObjectIds.NumericRange))
        self.add_property('OldValue', None, ua.VariantType.Variant)
        self.add_property('NewValue', None, ua.VariantType.Variant)


class AuditHistoryUpdateEvent(AuditUpdateEvent):
    """
    AuditHistoryUpdateEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditHistoryUpdateEventType)
        self.add_property('ParameterDataTypeId', ua.NodeId(ua.ObjectIds.AuditHistoryUpdateEventType), ua.VariantType.NodeId)


class AuditUpdateMethodEvent(AuditEvent):
    """
    AuditUpdateMethodEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditUpdateMethodEventType)
        self.add_property('MethodId', ua.NodeId(ua.ObjectIds.AuditUpdateMethodEventType), ua.VariantType.NodeId)
        self.add_property('InputArguments', None, ua.VariantType.Variant)


class SystemEvent(BaseEvent):
    """
    SystemEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.SystemEventType)


class DeviceFailureEvent(SystemEvent):
    """
    DeviceFailureEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.DeviceFailureEventType)


class BaseModelChangeEvent(BaseEvent):
    """
    BaseModelChangeEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.BaseModelChangeEventType)


class GeneralModelChangeEvent(BaseModelChangeEvent):
    """
    GeneralModelChangeEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.GeneralModelChangeEventType)
        self.add_property('Changes', None, ua.NodeId(ua.ObjectIds.ModelChangeStructureDataType))


class TransitionEvent(BaseEvent):
    """
    TransitionEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.TransitionEventType)
        self.add_property('Transition/Id', None, ua.VariantType.Variant)
        self.add_variable('Transition', None, ua.VariantType.LocalizedText)
        self.add_property('FromState/Id', None, ua.VariantType.Variant)
        self.add_variable('FromState', None, ua.VariantType.LocalizedText)
        self.add_property('ToState/Id', None, ua.VariantType.Variant)
        self.add_variable('ToState', None, ua.VariantType.LocalizedText)


class AuditUpdateStateEvent(AuditUpdateMethodEvent):
    """
    AuditUpdateStateEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditUpdateStateEventType)
        self.add_property('OldStateId', None, ua.VariantType.Variant)
        self.add_property('NewStateId', None, ua.VariantType.Variant)


class ProgramTransitionEvent(TransitionEvent):
    """
    ProgramTransitionEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.ProgramTransitionEventType)
        self.add_variable('IntermediateResult', None, ua.VariantType.Variant)


class SemanticChangeEvent(BaseEvent):
    """
    SemanticChangeEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.SemanticChangeEventType)
        self.add_property('Changes', None, ua.NodeId(ua.ObjectIds.SemanticChangeStructureDataType))


class AuditUrlMismatchEvent(AuditCreateSessionEvent):
    """
    AuditUrlMismatchEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditUrlMismatchEventType)
        self.add_property('EndpointUrl', None, ua.VariantType.String)


class Condition(BaseEvent):
    """
    Condition:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.ConditionType)
        self.add_property('ConditionClassId', ua.NodeId(ua.ObjectIds.ConditionType), ua.VariantType.NodeId)
        self.add_property('ConditionClassName', None, ua.VariantType.LocalizedText)
        self.add_property('ConditionSubClassId', ua.NodeId(ua.ObjectIds.ConditionType), ua.VariantType.NodeId)
        self.add_property('ConditionSubClassName', None, ua.VariantType.LocalizedText)
        self.add_property('ConditionName', None, ua.VariantType.String)
        self.add_property('BranchId', ua.NodeId(ua.ObjectIds.ConditionType), ua.VariantType.NodeId)
        self.add_property('Retain', None, ua.VariantType.Boolean)
        self.add_property('EnabledState/Id', None, ua.VariantType.Boolean)
        self.add_property('EnabledState/EffectiveDisplayName', None, ua.VariantType.LocalizedText)
        self.add_property('EnabledState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('EnabledState/EffectiveTransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('EnabledState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('EnabledState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('EnabledState', None, ua.VariantType.LocalizedText)
        self.add_property('Quality/SourceTimestamp', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_variable('Quality', None, ua.VariantType.StatusCode)
        self.add_property('LastSeverity/SourceTimestamp', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_variable('LastSeverity', None, ua.VariantType.UInt16)
        self.add_property('Comment/SourceTimestamp', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_variable('Comment', None, ua.VariantType.LocalizedText)
        self.add_property('ClientUserId', None, ua.VariantType.String)


class RefreshStartEvent(SystemEvent):
    """
    RefreshStartEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.RefreshStartEventType)


class RefreshEndEvent(SystemEvent):
    """
    RefreshEndEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.RefreshEndEventType)


class RefreshRequiredEvent(SystemEvent):
    """
    RefreshRequiredEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.RefreshRequiredEventType)


class AuditConditionEvent(AuditUpdateMethodEvent):
    """
    AuditConditionEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionEventType)


class AuditConditionEnableEvent(AuditConditionEvent):
    """
    AuditConditionEnableEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionEnableEventType)


class AuditConditionCommentEvent(AuditConditionEvent):
    """
    AuditConditionCommentEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionCommentEventType)
        self.add_property('ConditionEventId', None, ua.VariantType.ByteString)
        self.add_property('Comment', None, ua.VariantType.LocalizedText)


class DialogCondition(Condition):
    """
    DialogCondition:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.DialogConditionType)
        self.add_property('EnabledState/Id', None, ua.VariantType.Boolean)
        self.add_variable('EnabledState', None, ua.VariantType.LocalizedText)
        self.add_property('DialogState/Id', None, ua.VariantType.Boolean)
        self.add_property('DialogState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('DialogState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('DialogState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('DialogState', None, ua.VariantType.LocalizedText)
        self.add_property('Prompt', None, ua.VariantType.LocalizedText)
        self.add_property('ResponseOptionSet', None, ua.VariantType.LocalizedText)
        self.add_property('DefaultResponse', None, ua.VariantType.Int32)
        self.add_property('OkResponse', None, ua.VariantType.Int32)
        self.add_property('CancelResponse', None, ua.VariantType.Int32)
        self.add_property('LastResponse', None, ua.VariantType.Int32)


class AcknowledgeableCondition(Condition):
    """
    AcknowledgeableCondition:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AcknowledgeableConditionType)
        self.add_property('EnabledState/Id', None, ua.VariantType.Boolean)
        self.add_variable('EnabledState', None, ua.VariantType.LocalizedText)
        self.add_property('AckedState/Id', None, ua.VariantType.Boolean)
        self.add_property('AckedState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('AckedState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('AckedState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('AckedState', None, ua.VariantType.LocalizedText)
        self.add_property('ConfirmedState/Id', None, ua.VariantType.Boolean)
        self.add_property('ConfirmedState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('ConfirmedState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('ConfirmedState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('ConfirmedState', None, ua.VariantType.LocalizedText)


class AlarmCondition(AcknowledgeableCondition):
    """
    AlarmCondition:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AlarmConditionType)
        self.add_property('EnabledState/Id', None, ua.VariantType.Boolean)
        self.add_variable('EnabledState', None, ua.VariantType.LocalizedText)
        self.add_property('ActiveState/Id', None, ua.VariantType.Boolean)
        self.add_property('ActiveState/EffectiveDisplayName', None, ua.VariantType.LocalizedText)
        self.add_property('ActiveState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('ActiveState/EffectiveTransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('ActiveState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('ActiveState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('ActiveState', None, ua.VariantType.LocalizedText)
        self.add_property('InputNode', ua.NodeId(ua.ObjectIds.AlarmConditionType), ua.VariantType.NodeId)
        self.add_property('SuppressedState/Id', None, ua.VariantType.Boolean)
        self.add_property('SuppressedState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('SuppressedState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('SuppressedState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('SuppressedState', None, ua.VariantType.LocalizedText)
        self.add_property('OutOfServiceState/Id', None, ua.VariantType.Boolean)
        self.add_property('OutOfServiceState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('OutOfServiceState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('OutOfServiceState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('OutOfServiceState', None, ua.VariantType.LocalizedText)
        self.add_property('SuppressedOrShelved', None, ua.VariantType.Boolean)
        self.add_property('MaxTimeShelved', None, ua.NodeId(ua.ObjectIds.Duration))
        self.add_property('AudibleEnabled', None, ua.VariantType.Boolean)
        self.add_property('AudibleSound', None, ua.NodeId(ua.ObjectIds.AudioDataType))
        self.add_property('SilenceState/Id', None, ua.VariantType.Boolean)
        self.add_property('SilenceState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('SilenceState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('SilenceState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('SilenceState', None, ua.VariantType.LocalizedText)
        self.add_property('OnDelay', None, ua.NodeId(ua.ObjectIds.Duration))
        self.add_property('OffDelay', None, ua.NodeId(ua.ObjectIds.Duration))
        self.add_variable('FirstInGroupFlag', None, ua.VariantType.Boolean)
        self.add_property('LatchedState/Id', None, ua.VariantType.Boolean)
        self.add_property('LatchedState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('LatchedState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('LatchedState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('LatchedState', None, ua.VariantType.LocalizedText)
        self.add_property('ReAlarmTime', None, ua.NodeId(ua.ObjectIds.Duration))
        self.add_variable('ReAlarmRepeatCount', None, ua.VariantType.Int16)


class LimitAlarm(AlarmCondition):
    """
    LimitAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.LimitAlarmType)
        self.add_property('HighHighLimit', None, ua.VariantType.Double)
        self.add_property('HighLimit', None, ua.VariantType.Double)
        self.add_property('LowLimit', None, ua.VariantType.Double)
        self.add_property('LowLowLimit', None, ua.VariantType.Double)
        self.add_property('BaseHighHighLimit', None, ua.VariantType.Double)
        self.add_property('BaseHighLimit', None, ua.VariantType.Double)
        self.add_property('BaseLowLimit', None, ua.VariantType.Double)
        self.add_property('BaseLowLowLimit', None, ua.VariantType.Double)


class AuditHistoryEventUpdateEvent(AuditHistoryUpdateEvent):
    """
    AuditHistoryEventUpdateEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditHistoryEventUpdateEventType)
        self.add_property('UpdatedNode', ua.NodeId(ua.ObjectIds.AuditHistoryEventUpdateEventType), ua.VariantType.NodeId)
        self.add_property('PerformInsertReplace', None, ua.NodeId(ua.ObjectIds.PerformUpdateType))
        self.add_property('Filter', None, ua.NodeId(ua.ObjectIds.EventFilter))
        self.add_property('NewValues', None, ua.NodeId(ua.ObjectIds.HistoryEventFieldList))
        self.add_property('OldValues', None, ua.NodeId(ua.ObjectIds.HistoryEventFieldList))


class AuditHistoryValueUpdateEvent(AuditHistoryUpdateEvent):
    """
    AuditHistoryValueUpdateEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditHistoryValueUpdateEventType)
        self.add_property('UpdatedNode', ua.NodeId(ua.ObjectIds.AuditHistoryValueUpdateEventType), ua.VariantType.NodeId)
        self.add_property('PerformInsertReplace', None, ua.NodeId(ua.ObjectIds.PerformUpdateType))
        self.add_property('NewValues', None, ua.NodeId(ua.ObjectIds.DataValue))
        self.add_property('OldValues', None, ua.NodeId(ua.ObjectIds.DataValue))


class AuditHistoryDeleteEvent(AuditHistoryUpdateEvent):
    """
    AuditHistoryDeleteEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditHistoryDeleteEventType)
        self.add_property('UpdatedNode', ua.NodeId(ua.ObjectIds.AuditHistoryDeleteEventType), ua.VariantType.NodeId)


class AuditHistoryRawModifyDeleteEvent(AuditHistoryDeleteEvent):
    """
    AuditHistoryRawModifyDeleteEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditHistoryRawModifyDeleteEventType)
        self.add_property('IsDeleteModified', None, ua.VariantType.Boolean)
        self.add_property('StartTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('EndTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('OldValues', None, ua.NodeId(ua.ObjectIds.DataValue))


class AuditHistoryAtTimeDeleteEvent(AuditHistoryDeleteEvent):
    """
    AuditHistoryAtTimeDeleteEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditHistoryAtTimeDeleteEventType)
        self.add_property('ReqTimes', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('OldValues', None, ua.NodeId(ua.ObjectIds.DataValue))


class AuditHistoryEventDeleteEvent(AuditHistoryDeleteEvent):
    """
    AuditHistoryEventDeleteEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditHistoryEventDeleteEventType)
        self.add_property('EventIds', None, ua.VariantType.ByteString)
        self.add_property('OldValues', None, ua.NodeId(ua.ObjectIds.HistoryEventFieldList))


class EventQueueOverflowEvent(BaseEvent):
    """
    EventQueueOverflowEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.EventQueueOverflowEventType)


class ProgramTransitionAuditEvent(AuditUpdateStateEvent):
    """
    ProgramTransitionAuditEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.ProgramTransitionAuditEventType)
        self.add_property('Transition/Id', ua.NodeId(ua.ObjectIds.ProgramTransitionAuditEventType), ua.VariantType.NodeId)
        self.add_variable('Transition', None, ua.VariantType.LocalizedText)


class AuditConditionRespondEvent(AuditConditionEvent):
    """
    AuditConditionRespondEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionRespondEventType)
        self.add_property('SelectedResponse', None, ua.VariantType.UInt32)


class AuditConditionAcknowledgeEvent(AuditConditionEvent):
    """
    AuditConditionAcknowledgeEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionAcknowledgeEventType)
        self.add_property('ConditionEventId', None, ua.VariantType.ByteString)
        self.add_property('Comment', None, ua.VariantType.LocalizedText)


class AuditConditionConfirmEvent(AuditConditionEvent):
    """
    AuditConditionConfirmEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionConfirmEventType)
        self.add_property('ConditionEventId', None, ua.VariantType.ByteString)
        self.add_property('Comment', None, ua.VariantType.LocalizedText)


class ExclusiveLimitAlarm(LimitAlarm):
    """
    ExclusiveLimitAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.ExclusiveLimitAlarmType)
        self.add_property('ActiveState/Id', None, ua.VariantType.Boolean)
        self.add_variable('ActiveState', None, ua.VariantType.LocalizedText)


class ExclusiveLevelAlarm(ExclusiveLimitAlarm):
    """
    ExclusiveLevelAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.ExclusiveLevelAlarmType)


class ExclusiveRateOfChangeAlarm(ExclusiveLimitAlarm):
    """
    ExclusiveRateOfChangeAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.ExclusiveRateOfChangeAlarmType)
        self.add_property('EngineeringUnits', None, ua.NodeId(ua.ObjectIds.EUInformation))


class ExclusiveDeviationAlarm(ExclusiveLimitAlarm):
    """
    ExclusiveDeviationAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.ExclusiveDeviationAlarmType)
        self.add_property('SetpointNode', ua.NodeId(ua.ObjectIds.ExclusiveDeviationAlarmType), ua.VariantType.NodeId)
        self.add_property('BaseSetpointNode', ua.NodeId(ua.ObjectIds.ExclusiveDeviationAlarmType), ua.VariantType.NodeId)


class NonExclusiveLimitAlarm(LimitAlarm):
    """
    NonExclusiveLimitAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.NonExclusiveLimitAlarmType)
        self.add_property('ActiveState/Id', None, ua.VariantType.Boolean)
        self.add_variable('ActiveState', None, ua.VariantType.LocalizedText)
        self.add_property('HighHighState/Id', None, ua.VariantType.Boolean)
        self.add_property('HighHighState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('HighHighState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('HighHighState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('HighHighState', None, ua.VariantType.LocalizedText)
        self.add_property('HighState/Id', None, ua.VariantType.Boolean)
        self.add_property('HighState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('HighState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('HighState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('HighState', None, ua.VariantType.LocalizedText)
        self.add_property('LowState/Id', None, ua.VariantType.Boolean)
        self.add_property('LowState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('LowState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('LowState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('LowState', None, ua.VariantType.LocalizedText)
        self.add_property('LowLowState/Id', None, ua.VariantType.Boolean)
        self.add_property('LowLowState/TransitionTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('LowLowState/TrueState', None, ua.VariantType.LocalizedText)
        self.add_property('LowLowState/FalseState', None, ua.VariantType.LocalizedText)
        self.add_variable('LowLowState', None, ua.VariantType.LocalizedText)


class NonExclusiveLevelAlarm(NonExclusiveLimitAlarm):
    """
    NonExclusiveLevelAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.NonExclusiveLevelAlarmType)


class NonExclusiveRateOfChangeAlarm(NonExclusiveLimitAlarm):
    """
    NonExclusiveRateOfChangeAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.NonExclusiveRateOfChangeAlarmType)
        self.add_property('EngineeringUnits', None, ua.NodeId(ua.ObjectIds.EUInformation))


class NonExclusiveDeviationAlarm(NonExclusiveLimitAlarm):
    """
    NonExclusiveDeviationAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.NonExclusiveDeviationAlarmType)
        self.add_property('SetpointNode', ua.NodeId(ua.ObjectIds.NonExclusiveDeviationAlarmType), ua.VariantType.NodeId)
        self.add_property('BaseSetpointNode', ua.NodeId(ua.ObjectIds.NonExclusiveDeviationAlarmType), ua.VariantType.NodeId)


class DiscreteAlarm(AlarmCondition):
    """
    DiscreteAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.DiscreteAlarmType)


class OffNormalAlarm(DiscreteAlarm):
    """
    OffNormalAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.OffNormalAlarmType)
        self.add_property('NormalState', ua.NodeId(ua.ObjectIds.OffNormalAlarmType), ua.VariantType.NodeId)


class TripAlarm(OffNormalAlarm):
    """
    TripAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.TripAlarmType)


class AuditConditionShelvingEvent(AuditConditionEvent):
    """
    AuditConditionShelvingEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionShelvingEventType)
        self.add_property('ShelvingTime', None, ua.NodeId(ua.ObjectIds.Duration))


class ProgressEvent(BaseEvent):
    """
    ProgressEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.ProgressEventType)
        self.add_property('Context', None, ua.VariantType.Variant)
        self.add_property('Progress', None, ua.VariantType.UInt16)


class SystemStatusChangeEvent(SystemEvent):
    """
    SystemStatusChangeEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.SystemStatusChangeEventType)
        self.add_property('SystemState', None, ua.NodeId(ua.ObjectIds.ServerState))


class SystemOffNormalAlarm(OffNormalAlarm):
    """
    SystemOffNormalAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.SystemOffNormalAlarmType)


class AuditProgramTransitionEvent(AuditUpdateStateEvent):
    """
    AuditProgramTransitionEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditProgramTransitionEventType)
        self.add_property('TransitionNumber', None, ua.VariantType.UInt32)


class TrustListUpdatedAuditEvent(AuditUpdateMethodEvent):
    """
    TrustListUpdatedAuditEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.TrustListUpdatedAuditEventType)


class CertificateUpdatedAuditEvent(AuditUpdateMethodEvent):
    """
    CertificateUpdatedAuditEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.CertificateUpdatedAuditEventType)
        self.add_property('CertificateGroup', ua.NodeId(ua.ObjectIds.CertificateUpdatedAuditEventType), ua.VariantType.NodeId)
        self.add_property('CertificateType', ua.NodeId(ua.ObjectIds.CertificateUpdatedAuditEventType), ua.VariantType.NodeId)


class CertificateExpirationAlarm(SystemOffNormalAlarm):
    """
    CertificateExpirationAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.CertificateExpirationAlarmType)
        self.add_property('ExpirationDate', None, ua.VariantType.DateTime)
        self.add_property('ExpirationLimit', None, ua.NodeId(ua.ObjectIds.Duration))
        self.add_property('CertificateType', ua.NodeId(ua.ObjectIds.CertificateExpirationAlarmType), ua.VariantType.NodeId)
        self.add_property('Certificate', None, ua.VariantType.ByteString)


class AuditConditionResetEvent(AuditConditionEvent):
    """
    AuditConditionResetEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionResetEventType)


class PubSubStatusEvent(SystemEvent):
    """
    PubSubStatusEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.PubSubStatusEventType)
        self.add_property('ConnectionId', ua.NodeId(ua.ObjectIds.PubSubStatusEventType), ua.VariantType.NodeId)
        self.add_property('GroupId', ua.NodeId(ua.ObjectIds.PubSubStatusEventType), ua.VariantType.NodeId)
        self.add_property('State', None, ua.NodeId(ua.ObjectIds.PubSubState))


class PubSubTransportLimitsExceedEvent(PubSubStatusEvent):
    """
    PubSubTransportLimitsExceedEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.PubSubTransportLimitsExceedEventType)
        self.add_property('Actual', None, ua.VariantType.UInt32)
        self.add_property('Maximum', None, ua.VariantType.UInt32)


class PubSubCommunicationFailureEvent(PubSubStatusEvent):
    """
    PubSubCommunicationFailureEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.PubSubCommunicationFailureEventType)
        self.add_property('Error', None, ua.VariantType.StatusCode)


class DiscrepancyAlarm(AlarmCondition):
    """
    DiscrepancyAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.DiscrepancyAlarmType)
        self.add_property('TargetValueNode', ua.NodeId(ua.ObjectIds.DiscrepancyAlarmType), ua.VariantType.NodeId)
        self.add_property('ExpectedTime', None, ua.NodeId(ua.ObjectIds.Duration))
        self.add_property('Tolerance', None, ua.VariantType.Double)


class AuditConditionSuppressionEvent(AuditConditionEvent):
    """
    AuditConditionSuppressionEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionSuppressionEventType)


class AuditConditionSilenceEvent(AuditConditionEvent):
    """
    AuditConditionSilenceEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionSilenceEventType)


class AuditConditionOutOfServiceEvent(AuditConditionEvent):
    """
    AuditConditionOutOfServiceEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditConditionOutOfServiceEventType)


class RoleMappingRuleChangedAuditEvent(AuditUpdateMethodEvent):
    """
    RoleMappingRuleChangedAuditEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.RoleMappingRuleChangedAuditEventType)


class KeyCredentialAuditEvent(AuditUpdateMethodEvent):
    """
    KeyCredentialAuditEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.KeyCredentialAuditEventType)
        self.add_property('ResourceUri', None, ua.VariantType.String)


class KeyCredentialUpdatedAuditEvent(KeyCredentialAuditEvent):
    """
    KeyCredentialUpdatedAuditEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.KeyCredentialUpdatedAuditEventType)


class KeyCredentialDeletedAuditEvent(KeyCredentialAuditEvent):
    """
    KeyCredentialDeletedAuditEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.KeyCredentialDeletedAuditEventType)


class InstrumentDiagnosticAlarm(OffNormalAlarm):
    """
    InstrumentDiagnosticAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.InstrumentDiagnosticAlarmType)


class SystemDiagnosticAlarm(OffNormalAlarm):
    """
    SystemDiagnosticAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.SystemDiagnosticAlarmType)


class AuditHistoryAnnotationUpdateEvent(AuditHistoryUpdateEvent):
    """
    AuditHistoryAnnotationUpdateEvent:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.AuditHistoryAnnotationUpdateEventType)
        self.add_property('PerformInsertReplace', None, ua.NodeId(ua.ObjectIds.PerformUpdateType))
        self.add_property('NewValues', None, ua.NodeId(ua.ObjectIds.DataValue))
        self.add_property('OldValues', None, ua.NodeId(ua.ObjectIds.DataValue))


class TrustListOutOfDateAlarm(SystemOffNormalAlarm):
    """
    TrustListOutOfDateAlarm:
    """
    def __init__(self, sourcenode=None, message=None, severity=1):
        super().__init__(sourcenode, message, severity)
        self.EventType = ua.NodeId(ua.ObjectIds.TrustListOutOfDateAlarmType)
        self.add_property('TrustListId', ua.NodeId(ua.ObjectIds.TrustListOutOfDateAlarmType), ua.VariantType.NodeId)
        self.add_property('LastUpdateTime', None, ua.NodeId(ua.ObjectIds.UtcTime))
        self.add_property('UpdateFrequency', None, ua.NodeId(ua.ObjectIds.Duration))


IMPLEMENTED_EVENTS = {
    ua.ObjectIds.BaseEventType: BaseEvent,
    ua.ObjectIds.AuditEventType: AuditEvent,
    ua.ObjectIds.AuditSecurityEventType: AuditSecurityEvent,
    ua.ObjectIds.AuditChannelEventType: AuditChannelEvent,
    ua.ObjectIds.AuditOpenSecureChannelEventType: AuditOpenSecureChannelEvent,
    ua.ObjectIds.AuditSessionEventType: AuditSessionEvent,
    ua.ObjectIds.AuditCreateSessionEventType: AuditCreateSessionEvent,
    ua.ObjectIds.AuditActivateSessionEventType: AuditActivateSessionEvent,
    ua.ObjectIds.AuditCancelEventType: AuditCancelEvent,
    ua.ObjectIds.AuditCertificateEventType: AuditCertificateEvent,
    ua.ObjectIds.AuditCertificateDataMismatchEventType: AuditCertificateDataMismatchEvent,
    ua.ObjectIds.AuditCertificateExpiredEventType: AuditCertificateExpiredEvent,
    ua.ObjectIds.AuditCertificateInvalidEventType: AuditCertificateInvalidEvent,
    ua.ObjectIds.AuditCertificateUntrustedEventType: AuditCertificateUntrustedEvent,
    ua.ObjectIds.AuditCertificateRevokedEventType: AuditCertificateRevokedEvent,
    ua.ObjectIds.AuditCertificateMismatchEventType: AuditCertificateMismatchEvent,
    ua.ObjectIds.AuditNodeManagementEventType: AuditNodeManagementEvent,
    ua.ObjectIds.AuditAddNodesEventType: AuditAddNodesEvent,
    ua.ObjectIds.AuditDeleteNodesEventType: AuditDeleteNodesEvent,
    ua.ObjectIds.AuditAddReferencesEventType: AuditAddReferencesEvent,
    ua.ObjectIds.AuditDeleteReferencesEventType: AuditDeleteReferencesEvent,
    ua.ObjectIds.AuditUpdateEventType: AuditUpdateEvent,
    ua.ObjectIds.AuditWriteUpdateEventType: AuditWriteUpdateEvent,
    ua.ObjectIds.AuditHistoryUpdateEventType: AuditHistoryUpdateEvent,
    ua.ObjectIds.AuditUpdateMethodEventType: AuditUpdateMethodEvent,
    ua.ObjectIds.SystemEventType: SystemEvent,
    ua.ObjectIds.DeviceFailureEventType: DeviceFailureEvent,
    ua.ObjectIds.BaseModelChangeEventType: BaseModelChangeEvent,
    ua.ObjectIds.GeneralModelChangeEventType: GeneralModelChangeEvent,
    ua.ObjectIds.TransitionEventType: TransitionEvent,
    ua.ObjectIds.AuditUpdateStateEventType: AuditUpdateStateEvent,
    ua.ObjectIds.ProgramTransitionEventType: ProgramTransitionEvent,
    ua.ObjectIds.SemanticChangeEventType: SemanticChangeEvent,
    ua.ObjectIds.AuditUrlMismatchEventType: AuditUrlMismatchEvent,
    ua.ObjectIds.ConditionType: Condition,
    ua.ObjectIds.RefreshStartEventType: RefreshStartEvent,
    ua.ObjectIds.RefreshEndEventType: RefreshEndEvent,
    ua.ObjectIds.RefreshRequiredEventType: RefreshRequiredEvent,
    ua.ObjectIds.AuditConditionEventType: AuditConditionEvent,
    ua.ObjectIds.AuditConditionEnableEventType: AuditConditionEnableEvent,
    ua.ObjectIds.AuditConditionCommentEventType: AuditConditionCommentEvent,
    ua.ObjectIds.DialogConditionType: DialogCondition,
    ua.ObjectIds.AcknowledgeableConditionType: AcknowledgeableCondition,
    ua.ObjectIds.AlarmConditionType: AlarmCondition,
    ua.ObjectIds.LimitAlarmType: LimitAlarm,
    ua.ObjectIds.AuditHistoryEventUpdateEventType: AuditHistoryEventUpdateEvent,
    ua.ObjectIds.AuditHistoryValueUpdateEventType: AuditHistoryValueUpdateEvent,
    ua.ObjectIds.AuditHistoryDeleteEventType: AuditHistoryDeleteEvent,
    ua.ObjectIds.AuditHistoryRawModifyDeleteEventType: AuditHistoryRawModifyDeleteEvent,
    ua.ObjectIds.AuditHistoryAtTimeDeleteEventType: AuditHistoryAtTimeDeleteEvent,
    ua.ObjectIds.AuditHistoryEventDeleteEventType: AuditHistoryEventDeleteEvent,
    ua.ObjectIds.EventQueueOverflowEventType: EventQueueOverflowEvent,
    ua.ObjectIds.ProgramTransitionAuditEventType: ProgramTransitionAuditEvent,
    ua.ObjectIds.AuditConditionRespondEventType: AuditConditionRespondEvent,
    ua.ObjectIds.AuditConditionAcknowledgeEventType: AuditConditionAcknowledgeEvent,
    ua.ObjectIds.AuditConditionConfirmEventType: AuditConditionConfirmEvent,
    ua.ObjectIds.ExclusiveLimitAlarmType: ExclusiveLimitAlarm,
    ua.ObjectIds.ExclusiveLevelAlarmType: ExclusiveLevelAlarm,
    ua.ObjectIds.ExclusiveRateOfChangeAlarmType: ExclusiveRateOfChangeAlarm,
    ua.ObjectIds.ExclusiveDeviationAlarmType: ExclusiveDeviationAlarm,
    ua.ObjectIds.NonExclusiveLimitAlarmType: NonExclusiveLimitAlarm,
    ua.ObjectIds.NonExclusiveLevelAlarmType: NonExclusiveLevelAlarm,
    ua.ObjectIds.NonExclusiveRateOfChangeAlarmType: NonExclusiveRateOfChangeAlarm,
    ua.ObjectIds.NonExclusiveDeviationAlarmType: NonExclusiveDeviationAlarm,
    ua.ObjectIds.DiscreteAlarmType: DiscreteAlarm,
    ua.ObjectIds.OffNormalAlarmType: OffNormalAlarm,
    ua.ObjectIds.TripAlarmType: TripAlarm,
    ua.ObjectIds.AuditConditionShelvingEventType: AuditConditionShelvingEvent,
    ua.ObjectIds.ProgressEventType: ProgressEvent,
    ua.ObjectIds.SystemStatusChangeEventType: SystemStatusChangeEvent,
    ua.ObjectIds.SystemOffNormalAlarmType: SystemOffNormalAlarm,
    ua.ObjectIds.AuditProgramTransitionEventType: AuditProgramTransitionEvent,
    ua.ObjectIds.TrustListUpdatedAuditEventType: TrustListUpdatedAuditEvent,
    ua.ObjectIds.CertificateUpdatedAuditEventType: CertificateUpdatedAuditEvent,
    ua.ObjectIds.CertificateExpirationAlarmType: CertificateExpirationAlarm,
    ua.ObjectIds.AuditConditionResetEventType: AuditConditionResetEvent,
    ua.ObjectIds.PubSubStatusEventType: PubSubStatusEvent,
    ua.ObjectIds.PubSubTransportLimitsExceedEventType: PubSubTransportLimitsExceedEvent,
    ua.ObjectIds.PubSubCommunicationFailureEventType: PubSubCommunicationFailureEvent,
    ua.ObjectIds.DiscrepancyAlarmType: DiscrepancyAlarm,
    ua.ObjectIds.AuditConditionSuppressionEventType: AuditConditionSuppressionEvent,
    ua.ObjectIds.AuditConditionSilenceEventType: AuditConditionSilenceEvent,
    ua.ObjectIds.AuditConditionOutOfServiceEventType: AuditConditionOutOfServiceEvent,
    ua.ObjectIds.RoleMappingRuleChangedAuditEventType: RoleMappingRuleChangedAuditEvent,
    ua.ObjectIds.KeyCredentialAuditEventType: KeyCredentialAuditEvent,
    ua.ObjectIds.KeyCredentialUpdatedAuditEventType: KeyCredentialUpdatedAuditEvent,
    ua.ObjectIds.KeyCredentialDeletedAuditEventType: KeyCredentialDeletedAuditEvent,
    ua.ObjectIds.InstrumentDiagnosticAlarmType: InstrumentDiagnosticAlarm,
    ua.ObjectIds.SystemDiagnosticAlarmType: SystemDiagnosticAlarm,
    ua.ObjectIds.AuditHistoryAnnotationUpdateEventType: AuditHistoryAnnotationUpdateEvent,
    ua.ObjectIds.TrustListOutOfDateAlarmType: TrustListOutOfDateAlarm,
    }
